import celery.signals
import copy
import logging
import json
import time
import traceback

from celery import current_task
from celery.utils.log import get_task_logger
from functools import wraps
from urllib.parse import urlparse

from a2ml.api.a2ml import A2ML
from a2ml.api.a2ml_model import A2MLModel
from a2ml.api.a2ml_experiment import A2MLExperiment
from a2ml.api.a2ml_project import A2MLProject
from a2ml.api.model_review.model_helper import ModelHelper
from a2ml.api.model_review.model_review import ModelReview
from a2ml.api.stats.feature_divergence import FeatureDivergence
from a2ml.api.utils import dict_dig, merge_dicts
from a2ml.api.utils.json_utils import json_dumps_np
from a2ml.api.utils.context import Context
from a2ml.api.utils.s3_fsclient import S3FSClient, BotoClient
from a2ml.tasks_queue.config import Config

from .celery_app import celeryApp

PERSISTENT_DELIVERY_MODE = 2

task_config = Config()
logger = get_task_logger(__name__)

def _log(payload, level=logging.INFO):
    if (level == logging.DEBUG and task_config.debug) or level == logging.INFO:
        logger.info(str(payload))

def _exception_message_chain(e):
    if isinstance(e, Exception) and e.__cause__:
        return str(e) + ' caused by ' + _exception_message_chain(e.__cause__)
    else:
        return str(e)

def _exception_traces_chain(e):
    if isinstance(e, Exception) and e.__cause__:
        ''.join(traceback.format_tb(e.__traceback__)) + '\n\n' + _exception_traces_chain(e.__cause__)
    else:
        return ''.join(traceback.format_tb(e.__traceback__))

def send_result_to_hub(json_data):
    import amqp

    json_data = json_dumps_np(json_data)

    # _log('Send JSON data to Hub: ' + json_data, level=logging.DEBUG)

    o = urlparse(task_config.broker_url)

    connection_params = {
        'host': o.hostname,
        'port': o.port,
        'userid': o.username,
        'password': o.password,
        'virtual_host': o.path.replace('/', '')
    }

    with amqp.Connection(**connection_params) as c:
        channel = c.channel()
        channel.queue_declare(task_config.task_result_queue, durable=True, auto_delete=False)

        channel.basic_publish(
            amqp.Message(json_data, delivery_mode=PERSISTENT_DELIVERY_MODE),
            routing_key=task_config.task_result_queue
        )

def build_and_send_task_result_to_hub(args, status, result, traceback=None):
    if args:
        params = args[0]

        if params and params.get('hub_info', {}).get('cluster_task_id'):
            response = {
                'type': 'TaskResult',
                'provider': params.get('provider'), # provider == None -> means external provider
                'hub_info': params['hub_info'],
                'status': status,
                'runtime': time.time() - current_task.start_time,
                'result': result,
                'traceback': traceback,
            }

            send_result_to_hub(response)

@celery.signals.task_prerun.connect
def celery_task_prerun(**kwargs):
    current_task.start_time = time.time()

def process_task_result(task_func):
    @wraps(task_func)
    def func_wrapper(*args, **kwargs):
        try:
            result = task_func(*args, **kwargs)
            build_and_send_task_result_to_hub(args, 'success', result)
            return result
        except Exception as e:
            build_and_send_task_result_to_hub(
                args,
                'failure',
                _exception_message_chain(e),
                _exception_traces_chain(e)
            )

            # Reraise exception to signal that task is failed
            raise

    return func_wrapper

def _make_hub_provider_info_update(ctx, provider, hub_info):
    #project, project_file, experiment, experiment_session
    project_name = ctx.config.get("name", parts=ctx.config.parts_changes)
    project_file_dataset = ctx.config.get("dataset", parts=ctx.config.parts_changes)
    # project_file_validation_dataset = ctx.config.get("experiment/validation_dataset", parts=ctx.config.parts_changes)

    experiment_name = ctx.config.get('experiment/name', parts=ctx.config.parts_changes)
    experiment_session_id = ctx.config.get('experiment/run_id', parts=ctx.config.parts_changes)

    provider_info = {}

    if project_name:
        provider_info['project'] = {'name': project_name}

    if project_file_dataset:
        provider_info['project_file'] = {'url': project_file_dataset}

    # if project_file_validation_dataset:
    #     if provider_info.get('project_file'):
    #         provider_info['project_file']['validation_dataset'] = project_file_validation_dataset
    #     else:
    #         provider_info['project_file'] = {'validation_dataset': project_file_validation_dataset}

    if experiment_name:
        provider_info['experiment'] = {'name': experiment_name}

    if experiment_session_id:
        provider_info['experiment_session'] = {'id': experiment_session_id}

    return {
        'type': 'ProviderInfoUpdate',
        'hub_info': hub_info,
        'provider': provider,
        'provider_info': {
            provider: provider_info
        }
    }

def _read_hub_experiment_session(ctx, params):
    experiment_session = params['hub_info']['experiment_session']
    provider = params['provider']

    evaluation_options = experiment_session.get('model_settings', {}).get('evaluation_options')
    dataset_statistics = experiment_session.get('dataset_statistics', {}).get('stat_data', [])

    ctx.config.set('model_type',
        "classification" if evaluation_options.get('classification', True) else "regression")
    _get_options_from_dataset_statistic(ctx.config, dataset_statistics)

    ctx.config.set('experiment/validation_source', evaluation_options.get('test_data_path'))

    ctx.config.set('experiment/cross_validation_folds',
        evaluation_options.get('crossValidationFolds', 5))
    split_options = evaluation_options.get('splitOptions', {})
    if split_options.get('trainRatio'):
        ctx.config.set('experiment/validation_size', 1.0-float(split_options.get('trainRatio')))
    if split_options.get('shuffleData'):
        ctx.config.set('experiment/shuffle', split_options.get('shuffleData'))
    if split_options.get('timeseriesSplit'):
        ctx.config.set('experiment/timeseries_fold_split', split_options.get('timeseriesSplit'))

    ctx.config.set('experiment/max_total_time',
        evaluation_options.get('max_total_time_mins', 60))

    ctx.config.set('experiment/max_cores_per_trial',
        evaluation_options.get('cpu_per_mt_algorithm', 1))
    ctx.config.set('experiment/max_concurrent_trials',
        evaluation_options.get('trials_per_worker', 1))

    ctx.config.set('experiment/max_eval_time',
        evaluation_options.get('max_eval_time_mins', 6))
    ctx.config.set('experiment/max_n_trials',
        evaluation_options.get('max_n_trials', 100))
    ctx.config.set('experiment/use_ensemble',
        evaluation_options.get('use_ensemble', True))
    ctx.config.set('experiment/metric', evaluation_options.get('scoring'))

    if evaluation_options.get('algorithms_to_exlude'):
        ctx.config.set('experiment/blocked_models', evaluation_options.get('algorithms_to_exlude'))
    if evaluation_options.get('allowed_algorithms'):
        ctx.config.set('experiment/allowed_models', evaluation_options.get('allowed_algorithms'))
    if evaluation_options.get('exit_score'):
        ctx.config.set('experiment/exit_score', evaluation_options.get('exit_score'))
    if evaluation_options.get('score_top_count'):
        ctx.config.set('experiment/score_top_count', evaluation_options.get('score_top_count'))
    if evaluation_options.get('retrain_policy_type'):
        ctx.config.set('experiment/retrain_policy_type', evaluation_options.get('retrain_policy_type'))
    if evaluation_options.get('retrain_policy_value'):
        ctx.config.set('experiment/retrain_policy_value', evaluation_options.get('retrain_policy_value'))

    return ctx

def _update_hub_objects(ctx, provider, params):
    hub_objects_update = _make_hub_provider_info_update(ctx, params.get('provider'), params.get('hub_info'))
    send_result_to_hub(hub_objects_update)

    return hub_objects_update

def _get_options_from_dataset_statistic(config, stat_data):
    excluded_features = []
    target_feature = None

    categoricals = []
    label_encoded = []
    time_series = []
    date_time = []

    for item in stat_data:
        if item.get('isTarget'):
            target_feature = item['column_name']

        if not item.get('use') and not item.get('isTarget'):
            excluded_features.append(item['column_name'])

        if item.get('use') or item.get('isTarget'):
            if item['datatype'] == 'categorical':
                categoricals.append(item['column_name'])
            if item['datatype'] == 'hashing':
                categoricals.append(item['column_name'])
                label_encoded.append(item['column_name'])
            if item['datatype'] == 'timeseries':
                time_series.append(item['column_name'])
            if item['datatype'] == 'datetime':
                date_time.append(item['column_name'])

    if target_feature:
        config.set('target', target_feature)
    if excluded_features:
        config.set('exclude', excluded_features)

    if label_encoded:
        config.set('experiment/label_encoded', label_encoded)
    if categoricals:
        config.set('experiment/categoricals', categoricals)
    if date_time:
        config.set('experiment/date_time', date_time)
    if time_series:
        config.set('experiment/time_series', time_series)

def execute_task(task, params, wait_for_result=False, delay=0):
    if wait_for_result:
        task.apply(args=[params], countdown = delay)
    else:
        task.apply_async(args=[params], countdown = delay)

def _format_leaderboard_for_hub(leaderboard):
    formatted_leaderboard_list = []

    for item in leaderboard:
        obj = {}
        uid = item['uid']
        obj[uid] = item
        formatted_leaderboard_list.append(obj)

    return formatted_leaderboard_list

def _update_hub_leaderboad(params, leaderboard):
    from a2ml.api.auger.experiment import AugerExperiment

    ctx = Context(debug=task_config.debug)
    experiment = AugerExperiment(ctx)
    _log(leaderboard, level=logging.DEBUG)

    data = {
        'type': 'Leaderboard',
        'provider': params['provider'],
        'hub_info': params['hub_info'],
        'trials': _format_leaderboard_for_hub(leaderboard.get('trials', [])),
        'evaluate_status': leaderboard.get('evaluate_status'),
    }

    send_result_to_hub(data)

def _create_provider_context(params):
    provider = params.get('provider', 'auger')

    ctx = Context(
        name=provider,
        path=params.get('hub_info', {}).get('project_path'),
        debug=task_config.debug
    )

    if provider == 'auger':
        auger_creds = params.get('auger_creds', None)
        if auger_creds:
            ctx.credentials = auger_creds

    ctx.set_runs_on_server(True)
    ctx.config.set('providers', [provider])

    hub_info = params.get('hub_info', {})
    provider_info = params.get('provider_info', {}).get(provider, {})
    project_name = dict_dig(provider_info, 'project', 'name') or hub_info.get('project_name')
    experiment_name = dict_dig(provider_info, 'experiment', 'name') or hub_info.get('experiment_name')

    if project_name:
        ctx.config.set('name', project_name, provider)
    if experiment_name:
        ctx.config.set('experiment/name', experiment_name, provider)
    if provider_info.get('experiment_session', {}).get('id'):
        ctx.config.set('experiment/run_id', provider_info['experiment_session']['id'], provider)
    if hub_info.get('experiment_session') and hub_info.get('experiment_session').get('id'):
        ctx.config.set('experiment/hub_session_id', hub_info['experiment_session']['id'], provider)
        ctx.config.set('experiment/hub_experiment_id', hub_info['experiment_session'].get('experiment_id'), provider)

    return ctx

def _get_leaderboad(params):
    ctx = _create_provider_context(params)
    provider = params['provider']

    res = A2ML(ctx).evaluate()
    _log(res, level=logging.DEBUG)

    if res.get(provider, {}).get('result'):
        data = res[provider]['data']
        leaderboard = data['leaderboard']
        status = data['status']
        trials_count = data.get('trials_count', 0)

        trials = []

        for item in leaderboard:
            score = item['all_scores'].get(item['primary_metric'])
            if score is None:
                ctx.log('Leaderboard item does not contain primary score: %s'%item)
                score = 0.0

            trials.append({
                "uid": item['model id'],
                "score": score,
                "scoring": item['primary_metric'],
                "ensemble": 'Ensemble' in item['algorithm'],
                "task_type": item['task_type'],
                "all_scores": item['all_scores'],
                "score_name": item['primary_metric'],
                "algorithm_name": item['algorithm_name'],
                "optimizer_name": "Azure",
                "evaluation_time": float(item["fit_time"]),
                "algorithm_params": item['algorithm_params'],
                "experiment_session_id": ctx.config.get('experiment/run_id'),
                "preprocessor": item["preprocessor"],
                "algorithm_params_hash": None, #TODO : make_algorithm_params_hash in auger-ml

                "error": None,
                "ratio": 1.0,
                "budget": None,
                "create_trial_time": None,
                "estimated_time": 0,
                "estimated_timeout": False,
                "trialClass": None,
                "fold_scores": [],
                "fold_times": [],
                "metrics_time": 0,
                "dataset_ncols": 0,
                "dataset_nrows": 0,
                "dataset_manifest_id": None,
            })

        evaluate_status = {
            'status': status,
            'completed_evaluations': trials_count,
            'errors' : {'error': data.get('error'), 'error_details': data.get('error_details')}
        }

        return {
            'evaluate_status': evaluate_status,
            'trials': trials,
        }

@celeryApp.task(ignore_result=True)
@process_task_result
def monitor_evaluate_task(params):
    leaderboard = _get_leaderboad(params)

    if leaderboard:
        _update_hub_leaderboad(params, leaderboard)

    return True

@celeryApp.task(ignore_result=True)
@process_task_result
def evaluate_start_task(params):
    if not params.get('hub_info', {}).get('experiment_session_id'):
        raise Exception("evaluate_start_task missed experiment_session_id parameter.")

    ctx = _create_provider_context(params)
    provider = params.get('provider', 'auger')
    provider_info = params.get('provider_info', {}).get(provider, {})
    ctx.config.set('dataset', provider_info.get('project_file').get('url'), provider)

    cluster = provider_info.get('project', {}).get('cluster', {})
    ctx.config.set('cluster/name', cluster.get('name'), provider)
    ctx.config.set('cluster/min_nodes', cluster.get('min_nodes'), provider)
    ctx.config.set('cluster/max_nodes', cluster.get('max_nodes'), provider)
    ctx.config.set('cluster/type', cluster.get('type'), provider)

    ctx = _read_hub_experiment_session(ctx, params)
    ctx.config.clean_changes()

    res = A2ML(ctx).train()

    _update_hub_objects(ctx, provider, params)

    return res

@celeryApp.task(ignore_result=True)
@process_task_result
def stop_evaluate_task(params):
    ctx = _create_provider_context(params)

    ctx.config.clean_changes()
    res = A2MLExperiment(ctx).stop()
    _update_hub_objects(ctx, params.get('provider'), params)

    return res

@celeryApp.task(ignore_result=True)
@process_task_result
def update_cluster_config_task(params):
    ctx = _create_provider_context(params)
    clusters = params.get('clusters', [])
    for cluster in clusters:
        ctx.config.set('cluster/name', cluster.get('name'))
        res = A2MLProject(ctx).update_cluster_config(name=None, params=cluster)

    return True

@celeryApp.task(ignore_result=True)
@process_task_result
def import_data_task(params):
    if not params.get('hub_info').get('project_file'):
        raise Exception("import_data_task missed project_file parameter.")

    project_file = params['hub_info']['project_file']

    data_path = params.get('url')
    if not data_path:
        data_path = project_file.get('url')

    ctx = _create_provider_context(params)

    ctx.config.clean_changes()
    res = A2ML(ctx).import_data(source=data_path)
    _update_hub_objects(ctx, params.get('provider'), params)

    return res

@celeryApp.task(ignore_result=True)
@process_task_result
def deploy_model_task(params):
    ctx = _create_provider_context(params)
    provider = params.get('provider', 'auger')
    provider_info = params.get('provider_info', {}).get(provider, {})

    deploy_cluster = provider_info.get('project', {}).get('deploy_cluster', {})
    ctx.config.set('deploy_cluster/type', deploy_cluster.get('type'), provider)
    ctx.config.set('deploy_cluster/memory_gb', deploy_cluster.get('memory_gb'), provider)
    ctx.config.set('deploy_cluster/cpu_cores', deploy_cluster.get('cpu_cores'), provider)
    ctx.config.set('deploy_cluster/compute_target', deploy_cluster.get('compute_target'), provider)

    ctx = _read_hub_experiment_session(ctx, params)

    ctx.config.clean_changes()
    res = A2ML(ctx).deploy(model_id = params.get('model_id'), review = params.get('support_review_model'),
        locally=params.get('locally', False))
    _update_hub_objects(ctx, params.get('provider'), params)

    return res

@celeryApp.task(ignore_result=True)
@process_task_result
def add_external_model_task(params):
    return ModelReview(params).add_external_model(
        target_column=params.get('target_column'),
        scoring=params.get('scoring'),
        task_type=params.get('task_type'),
        binary_classification=params.get('binary_classification', False),
    )

@celeryApp.task(ignore_result=True)
@process_task_result
def undeploy_model_task(params):
    ctx = _create_provider_context(params)
    provider = params.get('provider', 'auger')

    model_id = params.get('hub_info', {}).get('pipeline_id')
    if not model_id:
        model_id = params.get('model_id')

    if not model_id:
        raise Exception("undeploy_model_task: hub_info/pipeline_id should be provided.")

    ctx.config.set('undeploy/service_only', params.get('service_only', False), provider)
    res = A2MLModel(ctx).undeploy(model_id = model_id, locally=params.get('locally', False))

    return res

@celeryApp.task(ignore_result=True)
@process_task_result
def predict_by_model_task(params):
    from a2ml.api.utils.crud_runner import CRUDRunner

    ctx = _create_provider_context(params)
    ctx = _read_hub_experiment_session(ctx, params)
    ctx.config.clean_changes()

    runner = CRUDRunner(ctx, "%s"%params.get('provider'), 'model')
    res = list(runner.providers.values())[0].predict(
        filename=params.get('path_to_predict'),
        model_id=params.get('model_id'),
        threshold=params.get('threshold'),
        data=params.get('records'),
        columns=params.get('features'),
        json_result=params.get('json_result'),
        count_in_result=params.get('count_in_result'),
        predicted_at=params.get('prediction_date'),
        prediction_id = params.get('prediction_id'),
        locally = params.get('locally', False),
        no_features_in_result=params.get('no_features_in_result', False)
    )
    _update_hub_objects(ctx, params.get('provider'), params)

    return res['predicted']

@celeryApp.task(ignore_result=True)
@process_task_result
def score_actuals_by_model_task(params):
    ctx = None
    external_model = params.get("external_model", False)

    if not external_model:
        ctx = _create_provider_context(params)
        ctx = _read_hub_experiment_session(ctx, params)
        ctx.config.clean_changes()

    return ModelReview(params).add_actuals(
        ctx,
        actuals_path = params.get('actuals_path'),
        data=params.get('actual_records'),
        columns=params.get('actual_columns', None),
        actual_date=params.get('actual_date'),
        actual_date_column=params.get('actual_date_column'),
        actuals_id=params.get('actuals_id'),
        return_count=params.get('return_count', False),
        provider=params.get('provider'),
        external_model=external_model,
    )

@celeryApp.task(ignore_result=True)
@process_task_result
def delete_actuals_task(params):
    return ModelReview(params).delete_actuals(
        with_predictions=params.get('with_predictions'),
        begin_date=params.get('begin_date'),
        end_date=params.get('end_date'),
    )

@celeryApp.task(ignore_result=True)
@process_task_result
def score_model_performance_daily_task(params):
    ctx = None
    external_model = params.get("external_model", False)

    if not external_model:
        ctx = _create_provider_context(params)
        ctx = _read_hub_experiment_session(ctx, params)
        ctx.config.clean_changes()

    return ModelReview(params).score_model_performance_daily(
        date_from=params.get('date_from'),
        date_to=params.get('date_to'),
        extra_features=params.get("features", []),
        provider=params.get('provider'),
        do_predict=params.get('do_predict'),
        ctx=ctx
    )

@celeryApp.task(ignore_result=True)
@process_task_result
def set_support_review_model_flag_task(params):
    return ModelReview(params).set_support_review_model_flag(
        flag_value=params.get('support_review_model')
    )

@celeryApp.task(ignore_result=True)
@process_task_result
def distribution_chart_stats_task(params):
    return ModelReview(params).distribution_chart_stats(
        date_from=params.get('date_from'),
        date_to=params.get('date_to'),
    )

@celeryApp.task(ignore_result=True)
@process_task_result
def remove_model_task(params):
    return ModelReview(params).remove_model()

@celeryApp.task(ignore_result=True)
@process_task_result
def clear_model_results_and_actuals(params):
    return ModelReview(params).clear_model_results_and_actuals()

@celeryApp.task(ignore_result=True)
@process_task_result
def build_review_data_task(params):
    ctx = _create_provider_context(params)
    ctx = _read_hub_experiment_session(ctx, params)

    date_col = None
    if ctx.config.get_list('experiment/date_time'):
        date_col = ctx.config.get_list('experiment/date_time')[0]

    retrain_policy = None
    if ctx.config.get('experiment/retrain_policy_type'):
        retrain_policy = {'type': ctx.config.get('experiment/retrain_policy_type'), 'value':ctx.config.get('experiment/retrain_policy_value')}

    return ModelReview(params).build_review_data(
        data_path=params.get('data_path'),
        date_col = date_col,
        retrain_policy = retrain_policy,
        date_to=params.get('date_to')
    )

@celeryApp.task(ignore_result=True)
@process_task_result
def build_divergence_model_task(params):
    return FeatureDivergence(params).build_and_save_model()

@celeryApp.task(ignore_result=True)
@process_task_result
def calc_divergence_daily_task(params):
    # In casse of ModelIsNotReadyError try to build it and retry 1 time
    for attempt in range(2):
        try:
            return FeatureDivergence(params).score_divergence_daily(
                date_from=params.get('date_from'),
                date_to=params.get('date_to'),
                top_n=params.get('top_n'),
            )
        except FeatureDivergence.ModelIsNotReadyError as e:
            if attempt == 0:
                FeatureDivergence(params).build_and_save_model()
            else:
                raise e

@celeryApp.task(ignore_result=True)
@process_task_result
def create_org_bucket(params):
    client = S3FSClient()
    client.ensure_bucket_created(
        Bucket=params['bucket_name'],
        region=params.get('region')
    )

    return True

@celeryApp.task(ignore_result=True)
@process_task_result
def delete_org_bucket(params):
    client = S3FSClient()
    client.ensure_bucket_deleted(Bucket=params['bucket_name'])

    return True

@celeryApp.task(ignore_result=True)
@process_task_result
def presign_s3_url_task(params):
    client = BotoClient(region=params.get('region'))

    path = params.get('path')
    if path:
        params['bucket'], params['key'] = S3FSClient.split_path_to_bucket_and_key(path)

    if params.get('multipart', False):
        return client.get_multipart_upload_config(
            bucket=params['bucket'],
            key=params['key'],
            expires_in=params.get('expires_in'),
        )
    else:
        return client.generate_presigned_url_ex(
            bucket=params['bucket'],
            key=params['key'],
            method=params['method'],
            expires_in=params.get('expires_in'),
            max_content_length=params.get('max_content_length'),
        )

@celeryApp.task(ignore_result=True)
@process_task_result
def validate_roi_syntax_task(params):
    return ModelReview(params).validate_roi_syntax(
        expressions=params["expressions"],
        features=params.get("features", []),
    )



# from auger.cli.cli import cli
# from auger.api.utils.config import Config
# from auger.api.cloud.project import AugerProjectApi

import pytest
import copy

from a2ml.api.auger.dataset import AugerDataset
from a2ml.api.auger.impl.cloud.project import AugerProjectApi
from .utils import interceptor, PAYLOAD_DEFAULT, object_status_chain


class TestDataSet():
    def test_list(self, log, project, ctx, authenticated, monkeypatch):
        PAYLOAD = PAYLOAD_DEFAULT
        interceptor(PAYLOAD, monkeypatch)

        result = AugerDataset(ctx).list()
        assert result.get('datasets')
        assert len(result['datasets']) == 2
        assert result['datasets'][0]['name'] == 'test_dataset1'
        assert result['datasets'][1]['name'] == 'test_dataset2'

    def test_create(self, log, project, ctx, authenticated, monkeypatch):
        PAYLOAD = copy.deepcopy(PAYLOAD_DEFAULT)
        PAYLOAD['get_project_files'] = {
            'data': [{
                'project_id': 1,
                'url': 's3://iris.csv',
                'name': 'iris-2.csv',
            }],
            'meta': {'pagination': {'offset': 0, 'total': 1, 'count': 1, 'limit': 100}, 'status': 200},
        }
        PAYLOAD['get_project_file'] = {'data': {}},
        PAYLOAD['create_project_file'] = {'data': {}}

        interceptor(PAYLOAD, monkeypatch)
        object_status_chain(['processing', 'processed'], monkeypatch)
        monkeypatch.setattr(AugerProjectApi, 'start', lambda self: None)
        monkeypatch.setattr('a2ml.api.auger.impl.cloud.dataset.AugerDataSetApi._upload_to_cloud', lambda *args: 's3://iris.csv')

        result = AugerDataset(ctx).create(source='iris.csv')
        assert result.get('created') == 'iris-1'

    def test_delete(self, log, project, ctx, authenticated, monkeypatch):
        PAYLOAD = copy.deepcopy(PAYLOAD_DEFAULT)
        PAYLOAD['delete_project_file'] = {'data': {}}

        interceptor(PAYLOAD, monkeypatch)

        result = AugerDataset(ctx).delete(name='test_dataset1')
        assert result.get('deleted') == 'test_dataset1'

    def test_select(self, log, project, ctx, authenticated):
        result = AugerDataset(ctx).select(name='iris.csv')
        assert result.get('selected') == 'iris.csv'
        print(result)



from a2ml.api.roi.base_interpreter import BaseInterpreter
from a2ml.api.roi.lexer import AstError, Lexer
from a2ml.api.roi.parser import Parser, TopNode

class ValidationError(AstError):
    pass

class ValidationWarning(AstError):
    pass

class ValidationResult:
    def __init__(self, is_valid=True, error=None, warning=None, tree=None):
        self.is_valid = is_valid
        self.error = error
        self.warning = warning
        self.tree = tree

    def __str__(self):
        return self.error or self.warning or ''

class Validator(BaseInterpreter):
    def __init__(self, expression, known_vars):
        self.expression = expression
        self.known_vars = known_vars
        self.known_funcs = BaseInterpreter.known_funcs()
        self.root = None

    def validate(self, force_raise=True):
        try:
            parser = Parser(Lexer(self.expression))
            self.root = parser.parse()
            self.evaluate(self.root)
            return ValidationResult(is_valid=True, tree=self.root)
        except ValidationWarning as e:
            return ValidationResult(is_valid=True, warning=str(e), tree=self.root)
        except AstError as e:
            if force_raise:
                raise e
            else:
                return ValidationResult(is_valid=False, error=str(e), tree=self.root)

    def evaluate_no_op_node(self, node):
        return True

    def evaluate_const_node(self, node):
        return True

    def evaluate_var_node(self, node):
        if node.name in self.known_vars:
            return True
        else:
            msg = f"unknown variable '{node.name}' at position {node.position()}"

            if node.name.startswith('$'):
                raise ValidationWarning(msg)
            else:
                raise ValidationError(msg)

    def evaluate_var_def_node(self, node):
        if node.name in self.known_vars:
            msg = f"var definition '{node.name}' at position {node.position()} conflicts with existing variable"
            raise ValidationError(msg)
        else:
            if isinstance(self.known_vars, set):
                self.known_vars.add(node.name)
            else:
                self.known_vars.append(node.name)

            return True

    def evaluate_binary_op_node(self, node):
        return self.evaluate(node.left) and self.evaluate(node.right)

    def evaluate_unary_op_node(self, node):
        return self.evaluate(node.node)

    def evaluate_func_node(self, node):
        if node.func_name in self.known_funcs:
            args_count = len(node.arg_nodes)

            if args_count in self.known_funcs[node.func_name]:
                return all(map(lambda node: self.evaluate(node), node.arg_nodes))
            else:
                expected_args_count = self.known_funcs[node.func_name]
                if isinstance(expected_args_count, range):
                    expected = f"from {expected_args_count.start} to {expected_args_count.stop - 1}"
                else:
                    counts = list(map(str, ))

                    if len(counts) > 1:
                        counts = counts[0:-2] + [counts[-2] + " or " + counts[-1]]

                    expected = ", ".join(counts)

                raise ValidationError(
                    f"invalid arguments count on '{node.func_name}' function, expected {expected} got {args_count} at position {node.position()}"
                )
        else:
            raise ValidationError(f"unknown function '{node.func_name}' at position {node.position()}")

    def evaluate_top_node(self, node):
        if not isinstance(self.root, TopNode):
            raise ValidationError(f"top or bottom expression cannot be used as an argument or operand")

        if node.with_node and not node.group_node:
            raise ValidationError(
                f"with experession at position {node.with_node.position()} can only be used with per expression"
            )

        return all(map(lambda n: self.evaluate(n), node.child_nodes()))

    def evaluate_with_node(self, node):
        return all(map(lambda n: self.evaluate(n), node.with_item_nodes))

    def evaluate_with_item_node(self, node):
        nodes = [node.source_node]

        if node.alias_node:
            nodes.append(node.alias_node)

        return all(map(lambda n: self.evaluate(n), nodes))



import os
import pandas as pd
import datetime
import math
import copy
import logging

from a2ml.api.utils import get_uid, convert_to_date, merge_dicts, fsclient
from a2ml.api.utils.dataframe import DataFrame
from a2ml.api.a2ml import A2ML, Context
from a2ml.api.roi.calculator import Calculator as RoiCalculator
from a2ml.api.roi.validator import Validator as RoiValidator
from a2ml.api.roi.validator import ValidationResult as RoiValidationResult
from a2ml.api.roi.interpreter import Interpreter as RoiInterpreter

from .model_helper import ModelHelper
from .probabilistic_counter import ProbabilisticCounter


class ModelReview(object):
    def __init__(self, params):
        self.model_id = params.get('hub_info', {}).get('pipeline_id')
        self.model_path = params.get('model_path')
        self.params = params

        if not self.model_path:
            self.model_path = ModelHelper.get_model_path(self.model_id, params['hub_info'].get('project_path'))

        self._load_options()


    # def get_actuals_statistic(self, date_from=None, date_to=None):
    #     count_actuals = self.count_actuals_by_prediction_id()
    #     performance_daily = self.score_model_performance_daily(date_from, date_to)
    #     distribution_chart_stats = self.distribution_chart_stats(date_from, date_to)

    #     return {
    #         'count_actuals': count_actuals,
    #         'performance_daily': performance_daily,
    #         'distribution_chart_stats': distribution_chart_stats
    #     }

    def _load_options(self):
        self.options_path = os.path.join(self.model_path, "options.json")
        self.options_file_exists = fsclient.is_file_exists(self.options_path)

        self.options = fsclient.read_json_file(self.options_path)

        if self.params.get('hub_info'):
            self.options['hub_info'] = self.params['hub_info']

        self.target_feature = self.options.get('targetFeature')
        self.original_features = self.options.get("originalFeatureColumns", [])


    def _do_score_actual(self, df_data, predicted_feature=None, extra_features=[]):
        ds_true = DataFrame({})

        ds_predict = DataFrame({})
        if predicted_feature:
            df_predicted_actual = df_data[[predicted_feature, 'a2ml_actual']].dropna()

            ds_true.df = df_predicted_actual.rename(columns={'a2ml_actual': self.target_feature})
            ds_predict.df = df_predicted_actual.rename(columns={predicted_feature: self.target_feature})
        else:
            ds_true.df = df_data[['a2ml_actual']].rename(columns={'a2ml_actual': self.target_feature})
            ds_predict.df = df_data[[self.target_feature]] # copy to prevent source data modification

        y_pred, _ = ModelHelper.preprocess_target_ds(self.model_path, ds_predict)
        y_true, _ = ModelHelper.preprocess_target_ds(self.model_path, ds_true)

        res = ModelHelper.calculate_scores(self.options, y_test=y_true, y_pred=y_pred, raise_main_score=False)

        roi = self._calculate_roi(df_data, predicted_feature, extra_features)
        if roi != None:
            res['roi'] = roi

        return res

    def validate_roi_syntax(self, expressions, features=[]):
        res = []
        known_vars = ["A", "P", "$" + self.target_feature] + list(
            map(lambda name: "$" + name, set(self.original_features + features))
        )

        for expression in expressions:
            if len(expression) > 0:
                validator = RoiValidator(expression, known_vars=known_vars)
                validation_result = validator.validate(force_raise=False)
            else:
                validation_result = RoiValidationResult()

            res.append(
                {
                    "expression": expression,
                    "is_valid": validation_result.is_valid,
                    "error": validation_result.error,
                    "warning": validation_result.warning,
                }
            )

        return res

    def _calculate_roi(self, df_data, predicted_feature=None, extra_features=[]):
        if not self.params.get('roi'):
            return None

        predicted_feature = predicted_feature or self.target_feature
        known_vars = list(set([predicted_feature] + self.original_features + extra_features))

        vars_mapping = {
            "A": "a2ml_actual",
            "actual": "a2ml_actual",
            "P": predicted_feature,
            "prediction": predicted_feature,
        }

        for known_var in known_vars:
            vars_mapping["$" + known_var] = known_var

        calc = RoiCalculator(
            filter=self.params['roi']['filter'],
            revenue=self.params['roi']['revenue'],
            investment=self.params['roi']['investment'],
            known_vars=known_vars,
            vars_mapping=vars_mapping,
        )

        res = calc.calculate(df_data)
        return res["roi"]

    def add_external_model(self, target_column, scoring, task_type, binary_classification):
        ModelHelper.create_model_options_file(
            options_path=self.options_path,
            scoring=scoring,
            target_column=target_column,
            task_type=task_type,
            binary_classification=binary_classification,
        )

        self._load_options()

        return True

    def add_actuals(
        self, ctx, actuals_path=None, data=None, columns=None, external_model=False,
        actual_date=None, actual_date_column=None, actuals_id = None, return_count=False, provider='auger',
    ):
        ds_actuals = DataFrame.create_dataframe(actuals_path, data, features=columns)

        if external_model or self.options.get('external_model'):
            options = self.options.copy()

            if 'hub_info' in options:
                del options['hub_info']

            ModelHelper.update_model_options_file(
                options_path=self.options_path,
                options=options,
                ds_actuals=ds_actuals,
            )

            self._load_options()

        if not 'actual' in ds_actuals.columns:
            raise Exception("There is no 'actual' column in data")

        actuals_count = ds_actuals.count()
        ds_actuals.df.rename(columns={"actual": 'a2ml_actual'}, inplace=True)

        if provider is not None and not self.target_feature in ds_actuals.columns:
            logging.info("Actual data missing predicted value column: %s. Call predict with features from actual data: %s"%(self.target_feature, ds_actuals.columns))
            self._do_predict(ctx, ds_actuals, provider)

        result = self._do_score_actual(ds_actuals.df)
        baseline_score = {}
        if "baseline_target" in ds_actuals.columns:
            baseline_score = self._do_score_actual(ds_actuals.df, "baseline_target")

        #logging.info("Actual result: %s", result)
        ds_actuals.df = ds_actuals.df.rename(columns={self.target_feature: 'a2ml_predicted'})
        ds_actuals.df = ds_actuals.df.rename(columns={'a2ml_actual': self.target_feature})

        if not actuals_id:
            actuals_id = get_uid()

        suffix = "full_data"
        if len(ds_actuals.columns) == 2:
            suffix = "no_features_data"

        if actual_date_column:
            ds_actuals.df[actual_date_column] = ds_actuals.df[actual_date_column].fillna(datetime.date.today()).apply(pd.to_datetime).dt.date

            uniq_dates = ds_actuals.df[actual_date_column].unique()
            uniq_dates.sort()

            for actual_date in uniq_dates:
                file_name = str(actual_date) + '_' + actuals_id + "_" + suffix + ".feather.zstd"
                df = DataFrame.create_dataframe(records=ds_actuals.df[ds_actuals.df[actual_date_column] == actual_date])
                df.saveToFeatherFile(os.path.join(self.model_path, "predictions", file_name))
        else:
            file_name = str(actual_date or datetime.date.today()) + '_' + actuals_id + "_" + suffix + ".feather.zstd"
            ds_actuals.saveToFeatherFile(os.path.join(self.model_path, "predictions", file_name))

        if return_count:
            return {'score': result, 'count': actuals_count, 'baseline_score': baseline_score}
        else:
            return result

    def _do_predict(self, ctx, ds_actuals, provider):
        missing_features = set(self.original_features) - set(ds_actuals.columns)
        if len(missing_features) > 0:
            missing_features = ', '.join(sorted(list(missing_features)))
            raise Exception(f'Missing features to make prediction: {missing_features}. Please, provide target \'{self.target_feature}\' or all training features to run predict.')

        res = A2ML(ctx).predict(self.model_id, data=ds_actuals.df, provider=provider)

        if res['result']:
            ds_actuals.df[self.target_feature] = res['data']['predicted'][self.target_feature]
        else:
            raise Exception(res['data'])
            
    def delete_actuals(self, with_predictions=False, begin_date=None, end_date=None):
        if with_predictions and not begin_date and not end_date:
            self.clear_model_results_and_actuals()
        else:
            path_suffixes = ["actuals.feather.zstd", "data.feather.zstd"]

            if with_predictions:
                path_suffixes = [".feather.zstd"]

            for path_suffix in path_suffixes:
                for (curr_date, files) in ModelReview._prediction_files_by_day(self.model_path, begin_date, end_date, path_suffix):
                    for file in files:
                        path = file if type(file) == str else file['path']
                        fsclient.remove_file(path)

    def build_review_data(self, data_path=None, output=None, date_col=None, retrain_policy=None,
        date_to=None):
        if not data_path:
            data_path = self.options['data_path']

        train_features = ModelHelper.get_train_features(self.options)
        ds_train = DataFrame.create_dataframe(data_path, features=train_features)

        all_files = fsclient.list_folder(
            os.path.join(self.model_path, "predictions/*_full_data.feather.zstd"),
            wild=True, remove_folder_name=False, meta_info=True
        )

        all_files.sort(key=lambda f: f['path'][0:10], reverse=True)
        if date_to:
            new_files = []
            date_to_date = convert_to_date(date_to)
            for idx, file in enumerate(all_files):
                file_date = os.path.basename(file['path']).split("_")[0]
                if convert_to_date(file_date) > date_to_date:
                    continue
                new_files.append(file)
            all_files = new_files

        start_date = None
        if '_review_date_' in data_path:
            start = data_path.index('_review_date_')+len('_review_date_')
            end = data_path.index('_', start)
            start_date = convert_to_date(data_path[start:end])

        if not start_date and date_col and date_col in train_features:
            try:
                start_date = convert_to_date(ds_train.df[date_col].max())
                logging.info("build_review_data with date_col: %s = %s"%(date_col,start_date))
            except Exception as e:
                logging.error("Getting latest date from data path %s failed: %s"%(data_path,e))

        if start_date:
            new_files = []
            for idx, file in enumerate(all_files):
                file_date = os.path.basename(file['path']).split("_")[0]
                if convert_to_date(file_date) <= start_date:
                    continue

                new_files.append(file)

            all_files = new_files

        logging.info("build_review_data adding files: %s"%all_files)
        for (file, ds_actuals) in DataFrame.load_from_files(all_files):
            if not ds_actuals.df.empty:
                ds_actuals.drop(['a2ml_predicted'])
                ds_train.df = pd.concat([ds_train.df, ds_actuals.df], ignore_index=True)
                #ds_train.drop_duplicates()

        #ds_train.dropna()
        if retrain_policy:
            if retrain_policy.get('type') == 'days_limit' and date_col and date_col in train_features:
                start_date = convert_to_date(ds_train.df[date_col].max())
                end_date = start_date - datetime.timedelta(days=int(retrain_policy.get('value')))
                ds_train.df.query("%s>='%s'"%(date_col, end_date), inplace=True)

        if not output:
            directory = os.path.dirname(data_path)
            file_name = os.path.basename(data_path).split('_review_')[0]
            date_suffix = ""
            if date_to:
                date_suffix = "date_%s_"%date_to

            output = os.path.join(directory, file_name + "_review_%s%s.parquet"%(date_suffix, get_uid()))

        ds_train.saveToFile(output)
        return output

    # date_from..date_to inclusive
    def score_model_performance_daily(self, date_from, date_to, extra_features=[], 
        provider='auger', do_predict=False, ctx=None):
        #To support baseline_target
        features = None #[self.target_feature, 'a2ml_predicted']
        res = {}

        for (curr_date, files) in ModelReview._prediction_files_by_day(
                self.model_path, date_from, date_to, "*_data.feather.zstd"):
            df_actuals = DataFrame({})
            for (file, df) in DataFrame.load_from_files(files, features):
                df_actuals.df = pd.concat([df_actuals.df, df.df])

            if df_actuals.count() > 0:
                #print(df_actuals.df['a2ml_predicted']) #.query("%s>=0.10"%'a2ml_predicted'))
                if do_predict:
                    self._do_predict(ctx, df_actuals, provider)

                df_actuals.df.rename(columns={self.target_feature: 'a2ml_actual'}, inplace=True)
                df_actuals.df.rename(columns={'a2ml_predicted': self.target_feature}, inplace=True)

                scores = self._do_score_actual(df_actuals.df, extra_features=extra_features)

                baseline_score = {}
                if "baseline_target" in df_actuals.columns:
                    baseline_score = self._do_score_actual(df_actuals.df, "baseline_target", extra_features)

                res[str(curr_date)] = {
                    'scores': scores,
                    'score_name': self.options.get('score_name'),
                    'baseline_scores': baseline_score
                }

        return res

    def distribution_chart_stats(self, date_from, date_to):
        features = [self.target_feature, 'a2ml_predicted']
        features += self.options.get('originalFeatureColumns', [])
        categoricalFeatures = self.options.get('categoricalFeatures', [])
        mapper = {}
        mapper[self.target_feature] = 'actual_%s' % self.target_feature
        mapper['a2ml_predicted'] = 'predicted_%s' % self.target_feature

        actuals_stats = self._distribution_stats(
            date_from, date_to, "_*_data.feather.zstd", features, categoricalFeatures, mapper
        )

        if not self.options.get('data_path'):
            all_files = []
            date_stat = convert_to_date(date_to) - datetime.timedelta(days=1)
            for (curr_date, files) in ModelReview._prediction_files_by_day(self.model_path, None,
                date_stat, "_*_data.feather.zstd"):
                all_files += files

            base_stat = ModelReview._get_distribution_stats_files(all_files, features, categoricalFeatures, mapper)
            if base_stat:
                actuals_stats['base_stat'] = base_stat

        return actuals_stats

    def set_support_review_model_flag(self, flag_value):
        path = os.path.join(self.model_path, "options.json")

        self.options['support_review_model'] = bool(flag_value)

        fsclient.remove_file(path)
        fsclient.write_json_file(path, self.options, atomic=True)

    def remove_model(self):
        fsclient.remove_folder(self.model_path)
        return True

    def clear_model_results_and_actuals(self):
        fsclient.remove_folder(os.path.join(self.model_path, "predictions"))
        return True

    @staticmethod
    def _get_distribution_stats_files(files, features, categoricalFeatures=[], feature_mapper={},
        feature_importances={}):

        if not files:
            return None

        stats = {}

        for feature in features:
            stats[feature] = {
                'count': 0,
                'sum': 0,
                'sq_sum': 0,
                'dist': None,
                'imp': feature_importances.get(feature, 0)
            }

        df_list = []
        for (file, df) in DataFrame.load_from_files(files):
            df_list.append(df)

        # First pass: calc sum and count in each column for average
        for df in df_list:
            for feature in features:
                if feature in df.columns:
                    stats[feature]['count'] += df.df[feature].count()

                    if df.df[feature].dtype.name in ['category', 'string', 'object'] or \
                        feature in categoricalFeatures:
                        stats[feature]['dist'] = merge_dicts(
                            stats[feature]['dist'] or {},
                            dict(df.df[feature].value_counts()),
                            lambda v, ov: v + ov
                        )
                    else:
                        stats[feature]['sum'] += df.df[feature].sum()

        # Calc average
        for feature in features:
            if stats[feature]['count'] > 0 and stats[feature]['dist'] == None:
                stats[feature]['average'] = stats[feature]['sum'] / stats[feature]['count']

        # Second pass: sum of squares of value and average for std dev
        for df in df_list:
            for feature in features:
                if 'average' in stats[feature] and feature in df.columns:
                    avg = stats[feature]['average']
                    stats[feature]['sq_sum'] += ((df.df[feature] - avg)**2).sum()

        # Calc std dev
        return ModelReview._calc_stddev_for_features(stats, features, feature_mapper)

    def _distribution_stats(self, date_from, date_to, path_suffix, features,
        categoricalFeatures=[], feature_mapper={}):
        res = {}
        feature_importances = self.get_feature_importances()

        for (curr_date, files) in ModelReview._prediction_files_by_day(self.model_path, date_from, date_to, path_suffix):

            stats = ModelReview._get_distribution_stats_files(files, features, categoricalFeatures, feature_mapper, feature_importances)
            # Calc std dev
            if stats:
                res[str(curr_date)] = stats

        return res

    def get_feature_importances(self):
        cache_path = ModelHelper.get_metric_path(self.options)

        importance_data = None
        if cache_path:
            importance_data = fsclient.read_json_file(os.path.join(cache_path, "metrics.json")).get('feature_importance_data')
            if not importance_data:
                importance_data = fsclient.read_json_file(os.path.join(cache_path, "metric_names_feature_importance.json")).get('feature_importance_data')

        if importance_data:
            if importance_data.get('features_orig') and importance_data.get('scores_orig'):
                return dict(zip(importance_data['features_orig'], importance_data['scores_orig']))
            else:
                return dict(zip(importance_data['features'], importance_data['scores']))
        else:
            logging.warn("No feature importance in cache: for model %s" % (cache_path))
            return {}

    @staticmethod
    def _calc_stddev_for_features(stats, features, feature_mapper):
        res = {}

        for feature in features:
            count = stats[feature]['count']

            if count > 0:
                feature_key = feature_mapper.get(feature, feature)
                if 'average' in stats[feature]:
                    res[feature_key] = {
                        "avg": stats[feature]['average'],
                        "std_dev": math.sqrt(stats[feature]['sq_sum'] / (count - 1)) if count > 1 else 0
                    }
                else:
                    res[feature_key] = { 'dist': stats[feature]['dist'] }

                res[feature_key]['imp'] = round(stats[feature]['imp'], 6)

        return res

    @staticmethod
    def _get_first_date_from_files(path):
        all_files = fsclient.list_folder(path,
            wild=True, remove_folder_name=True, meta_info=False
        )
        first_date = None
        if all_files:
            all_files.sort(key=lambda f: f[0:10], reverse=False)
            idxDate = all_files[0].find("_")
            if idxDate:
                first_date = all_files[0][0:idxDate]

        return first_date

    @staticmethod
    def _prediction_files_by_day(model_path, date_from, date_to, path_suffix):
        if (date_from and not date_to):# or (not date_from and date_to):
            # TODO: list all files by suffix, sort them by prefix date and return range of files
            raise Exception("Arguments error: please provide both start and end dates or date_to only or do not pass any.")

        if date_from or date_to:
            if date_from:
                date_from = convert_to_date(date_from)
            else:
                date_from = ModelReview._get_first_date_from_files(
                    os.path.join(model_path, "predictions/*" + path_suffix))
                if not date_from:
                    return
                date_from = convert_to_date(date_from)

            curr_date = date_from
            date_to = convert_to_date(date_to)

            while curr_date <= date_to:
                path = os.path.join(model_path, "predictions/" + str(curr_date) + "*" + path_suffix)
                files = fsclient.list_folder(path, wild=True, remove_folder_name=False, meta_info=False)
                yield (curr_date, files)
                curr_date += datetime.timedelta(days=1)
        else:
            path = os.path.join(model_path, "predictions/" + "*" + path_suffix)
            files = fsclient.list_folder(path, wild=True, remove_folder_name=False, meta_info=False)
            yield ("today", files)

    @staticmethod
    def _remove_duplicates_by(df, column_name, counter):
        dup_flag_column_name = '__duplicate'
        df.df[dup_flag_column_name] = df.df[column_name]
        df.df[dup_flag_column_name] = df.df[dup_flag_column_name].apply(lambda uid: not counter.add(uid))

        df.df = df.df[df.df[dup_flag_column_name] == False]
        df.drop([dup_flag_column_name])

        return df

    @staticmethod
    def _get_prediction_files(model_path, prediction_group_id=None):
        predictions_path = os.path.join(model_path, "predictions/*_results.feather.zstd")

        if prediction_group_id:
            predictions_path = os.path.join(model_path, f"predictions/*_{prediction_group_id}_results.feather.zstd")

        files = fsclient.list_folder(predictions_path, wild=True, remove_folder_name=False, meta_info=True)
        files.sort(key=lambda f: f['last_modified'], reverse=True)

        if len(files) == 0:
            raise Exception('there is no prediction results for this model in ' + predictions_path)

        return files



import boto3
import botocore
import datetime
import json
import logging
import mimetypes
import os
import time
import uuid

from a2ml.api.utils import retry_helper
from dateutil.tz import tzutc
from urllib.parse import urlparse

AWS_S3_HOST = "s3.amazonaws.com"

def retry_handler(decorated):
    def wrapper(self, *args, **kwargs):
        return retry_helper(lambda: decorated(self, *args, **kwargs), ['InvalidAccessKeyId', 'NoSuchKey'])

    return wrapper

class BotoClient:
    def __init__(self, region=None, aws_role_arn=None, endpoint_url=None):
        self.endpoint_url = endpoint_url or os.environ.get('S3_ENDPOINT_URL')
        self.aws_role_arn = aws_role_arn or os.environ.get('AWS_ROLE_ARN')
        self.region = region
        self.client = self._build_client('s3')

    def _build_client(self, service_name):
        return boto3.client(
            service_name,
            endpoint_url=self.endpoint_url,
            config=boto3.session.Config(signature_version='s3v4'),
            region_name=self.region,
        )

    @retry_handler
    def get_waiter(self, *args, **kwargs):
        return self.client.get_waiter(*args, **kwargs)

    @retry_handler
    def list_objects_v2(self, *args, **kwargs):
        return self.client.list_objects_v2(*args, **kwargs)

    @retry_handler
    def put_object(self, *args, **kwargs):
        return self.client.put_object(*args, **kwargs)

    @retry_handler
    def head_bucket(self, *args, **kwargs):
        return self.client.head_bucket(*args, **kwargs)

    @retry_handler
    def create_bucket(self, *args, **kwargs):
        return self.client.create_bucket(*args, **kwargs)

    @retry_handler
    def delete_bucket(self, Bucket):
        self._delete_items(Bucket, "list_objects", "Contents")
        self._delete_items(Bucket, "list_object_versions", "Versions")
        self._delete_items(Bucket, "list_object_versions", "DeleteMarkers")

        self.client.delete_bucket(Bucket=Bucket)

    def _map_version_key(self, obj):
        res = {"Key": obj["Key"]}

        if "VersionId" in obj:
            res["VersionId"] = obj["VersionId"]

        return res

    def _delete_items(self, bucket, method_name, resource_name):
        batch_delete_limit = 1000 # delete_objects can't delete more that 1000 keys at once
        list_func = lambda: getattr(self.client, method_name)(Bucket=bucket, MaxKeys=batch_delete_limit).get(resource_name, [])

        items = list_func()
        while len(items) > 0:
            keys = list(map(self._map_version_key, items))
            self.client.delete_objects(Bucket=bucket, Delete={'Objects': keys})
            items = list_func()

    @retry_handler
    def delete_object(self, *args, **kwargs):
        return self.client.delete_object(*args, **kwargs)

    @retry_handler
    def copy_object(self, *args, **kwargs):
        return self.client.copy_object(*args, **kwargs)

    @retry_handler
    def head_object(self, *args, **kwargs):
        return self.client.head_object(*args, **kwargs)

    @retry_handler
    def list_objects(self, *args, **kwargs):
        return self.client.list_objects(*args, **kwargs)

    @retry_handler
    def download_file(self, *args, **kwargs):
        return self.client.download_file(*args, **kwargs)

    @retry_handler
    def upload_file(self, *args, **kwargs):
        return self.client.upload_file(*args, **kwargs)

    @retry_handler
    def copy(self, *args, **kwargs):
        return self.client.copy(*args, **kwargs)

    @retry_handler
    def generate_presigned_url_ex(self, bucket, key, method="GET", expires_in=None, max_content_length=None):
        response = self.client.get_bucket_location(Bucket=bucket)

        s3_client = boto3.client(
            's3',
            endpoint_url=self.endpoint_url,
            config=boto3.session.Config(
                signature_version='s3v4',
                region_name=response.get('LocationConstraint'),
                s3={'addressing_style': 'virtual'},
            )
        )

        if method == 'POST':
            conditions = [{"success_action_status": "200"}]

            if max_content_length:
                conditions.append(["content-length-range", 0, max_content_length])

            return s3_client.generate_presigned_post(
                Bucket=bucket,
                Key=key,
                ExpiresIn=expires_in or 3600,
                Fields={'success_action_status': 200},
                Conditions=conditions,
            )
        elif method == "GET" or method == "PUT":
            client_method = 'get_object'

            if method == 'PUT':
                client_method = 'put_object'

            return s3_client.generate_presigned_url(
                ClientMethod=client_method,
                ExpiresIn=expires_in,
                Params={
                    'Bucket': bucket,
                    'Key': key
                },
            )
        else:
            raise ValueError(f"unexpected method: '{method}'")

    @retry_handler
    def get_multipart_upload_config(self, bucket, key, expires_in=None):
        sts_client = self._build_client('sts')

        response = sts_client.assume_role(
            RoleArn=self.aws_role_arn,
            RoleSessionName='upload_' + str(uuid.uuid4()),
            DurationSeconds=expires_in,
            Policy=self._build_upload_polciy(bucket, key)
        )

        credentials = response["Credentials"]
        endpoint = self.client._endpoint.host
        parsed_url = urlparse(endpoint)

        # Multipart upload requires host if format: "Host: Bucket.s3.amazonaws.com"
        # see https://docs.aws.amazon.com/AmazonS3/latest/API/API_CreateMultipartUpload.html
        # but Minio requires plain URL, so check netloc and insert bucket only for AWS S3
        if parsed_url.netloc == AWS_S3_HOST:
            endpoint = None

        return {
            "bucket": bucket,
            "key": key,
            "config": {
                "endpoint": endpoint,
                "port": parsed_url.port or 443,
                "use_ssl": True,
                "access_key": credentials["AccessKeyId"],
                "secret_key": credentials["SecretAccessKey"],
                "security_token": credentials["SessionToken"],
            }
        }

    def _build_upload_polciy(self, bucket, key):
        return json.dumps(
            {
                "Version": '2012-10-17',
                "Statement": [
                    {
                        "Action": [
                            "s3:*",
                        ],
                        "Effect": "Allow",
                        "Resource": [
                            f"arn:aws:s3:::{bucket}/{key}"
                        ],
                    }
                ]
            }
        )

    def get_waiter_names(self):
        return self.client.waiter_names

class S3FSClient:
    @staticmethod
    def split_path_to_bucket_and_key(path):
        uri = urlparse(path)

        rel_path = uri.path[1:]
        rel_path = rel_path.replace("//", "/")
        if rel_path.endswith("/"):
            rel_path = rel_path[:-1]

        return uri.netloc, rel_path

    def _get_relative_path(self, path):
        self.s3BucketName, rel_path = S3FSClient.split_path_to_bucket_and_key(path)
        self.client = BotoClient()
        return rel_path

    def get_smart_open_transport_params(self):
        transport_params = None
        endpoint_url = os.environ.get('S3_ENDPOINT_URL')

        if endpoint_url:
            transport_params = {
                'resource_kwargs': {
                    'endpoint_url': endpoint_url,
                }
            }

        return transport_params

    def generate_presigned_url(self, path):
        try:
            path = self._get_relative_path(path)

            return self.client.generate_presigned_url_ex(self.s3BucketName, path)
        except Exception as e:
            logging.error("generate_presigned_url failed: %s"%e)

        return None

    def wait_for_path(self, path):
        path = self._get_relative_path(path)
        if 'object_exists' in self.client.get_waiter_names():
            waiter = self.client.get_waiter('object_exists')
            waiter.config.delay = 2
            waiter.config.max_attempts = 50

            logging.info("S3FSClient: waiting for max %s sec for object: %s" % (
                waiter.config.delay*waiter.config.max_attempts, path))
            waiter.wait(Bucket=self.s3BucketName, Key=path)

    def s3fs_open(self, path, mode):
        from s3fs.core import S3FileSystem

        endpoint_url = os.environ.get('S3_ENDPOINT_URL')
        client_kwargs = {}
        if endpoint_url:
            client_kwargs = {'endpoint_url': endpoint_url}

        if 'r' in mode:
            self.wait_for_path(path)

        s3 = S3FileSystem(anon=False, default_fill_cache=False,
                          client_kwargs=client_kwargs)
        return s3.open(path, mode=mode)

    def list_folder(self, path, wild=False, remove_folder_name=True, meta_info=False):
        import fnmatch
        from urllib.parse import unquote

        path_arg = path
        path = self._get_relative_path(path)
        path_filter = None
        if path and wild:
            parts = path.split('/')
            path = '/'.join(parts[:-1])
            path_filter = parts[-1]
            parts = path_arg.split('/')
            path_arg = '/'.join(parts[:-1])

        if path:
            path = path + "/" if not path.endswith("/") else path

        #print("s3BucketName: %s;Path:%s; path_filter:%s"%(self.s3BucketName, path, path_filter))

        ContinuationToken = None
        result = []
        while True:
            if ContinuationToken:
                listFiles = self.client.list_objects_v2(
                    Bucket=self.s3BucketName, Prefix=path, Delimiter='/', ContinuationToken=ContinuationToken)
            else:
                listFiles = self.client.list_objects_v2(
                    Bucket=self.s3BucketName, Prefix=path, Delimiter='/')

            ContinuationToken = listFiles.get('NextContinuationToken', None)
            files = None
            folders = None
            if listFiles is not None:
                files = listFiles.get('Contents')
                folders = listFiles.get('CommonPrefixes')

            if files is not None:
                for file in files:
                    if file.get('Key') != path:
                        file_path = unquote(file.get('Key')[len(path):])
                        if meta_info:
                            result.append({'path': file_path,
                                           'last_modified': self._get_seconds_from_epoch(file.get('LastModified')), 'size': file.get('Size', 0)})
                        else:
                            result.append(file_path)

            if folders is not None:
                for folder in folders:
                    if folder.get('Prefix') != path:
                        folder_path = unquote(folder.get('Prefix')[len(path):])
                        if meta_info:
                            result.append({'path': folder_path,
                                           'last_modified': self._get_seconds_from_epoch(folder.get('LastModified')), 'size': folder.get('Size', 0)})
                        else:
                            result.append(folder_path)

            if not ContinuationToken:
                break

        if path_filter:
            if meta_info:
                result = [item for item in result if fnmatch.fnmatch(
                    item['path'], path_filter)]
            else:
                result = fnmatch.filter(result, path_filter)

        if not remove_folder_name:
            for idx, item in enumerate(result):
                if meta_info:
                    result[idx]['path'] = os.path.join(path_arg, item['path'])
                else:
                    result[idx] = os.path.join(path_arg, result[idx])

        return result

    def create_folder(self, path):
        path = self._get_relative_path(path)
        path = path + "/" if not path.endswith("/") else path

        self.ensure_bucket_created(self.s3BucketName)

        # If no path we shouldn't create it
        if path != "/":
            self.client.put_object(Bucket=self.s3BucketName, Key=path.lstrip('/'))

    def create_parent_folder(self, path):
        parent = os.path.dirname(path)
        if parent:
            self.create_folder(parent)

    def ensure_bucket_created(self, Bucket, region=None):
        try:
            self.client = BotoClient(region=region)
            self.client.head_bucket(Bucket=Bucket)
        except botocore.client.ClientError as e:
            if e.response['Error']['Code'] == '404':
                kwargs = {'Bucket': Bucket}

                if region:
                    kwargs['CreateBucketConfiguration'] = {'LocationConstraint': region}

                self.client.create_bucket(**kwargs)
            else:
                raise

        try:
            config = {
                'CORSRules': [
                    {
                        "AllowedHeaders": ["*"],
                        "AllowedMethods": ["GET", "PUT", "POST", "DELETE"],
                        "AllowedOrigins": ["*"],
                        "ExposeHeaders": ["ETag", "x-amz-request-id", "x-amz-id-2"],
                        "MaxAgeSeconds": 3600,
                    }
                ]
            }

            self.client.client.put_bucket_cors(Bucket=Bucket, CORSConfiguration=config)
        except botocore.exceptions.ClientError as e:
            if not "NotImplemented" in str(e):
                raise
            else:
                pass
                # Minio doesn't support CORSes yet https://docs.min.io/docs/minio-server-limits-per-tenant.html

    def ensure_bucket_deleted(self, Bucket):
        try:
            self.client = BotoClient()
            self.client.head_bucket(Bucket=Bucket)
            self.client.delete_bucket(Bucket=Bucket)
        except botocore.client.ClientError as e:
            code = e.response['Error']['Code']
            # 404 - deleted, that how it should work
            # 400 - looks like a bug in S3 API, it's returned for deleted versioned buckets
            # same as in https://github.com/aws/aws-sdk-js/issues/2749
            if code == '404' or code == '400':
                return True
            else:
                raise


    def remove_folder(self, path, remove_self=True):
        path = self._get_relative_path(path)
        self._s3_removeFolder(path, remove_self)

    def remove_file(self, path, wild=False):
        try:
            if wild:
                files = self.list_folder(path, wild)
                path = self._get_relative_path(path)
                for file in files:
                    path_to_remove = os.path.join(os.path.dirname(path), file)
                    # print(path_to_remove)
                    self.client.delete_object(
                        Bucket=self.s3BucketName, Key=path_to_remove)

                #raise Exception("S3FSClient::remove_file wild is not implemented:%s"%path)
            else:
                path = self._get_relative_path(path)
                self.client.delete_object(Bucket=self.s3BucketName, Key=path)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] != 'NoSuchBucket':
                raise

    def _s3_removeFolder(self, path, remove_self=True):
        path = path + "/" if not path.endswith("/") else path

        # print("Remove:%s"%path)
        ContinuationToken = None
        while True:
            if ContinuationToken:
                listFiles = self.client.list_objects_v2(
                    Bucket=self.s3BucketName, Prefix=path, Delimiter='/', ContinuationToken=ContinuationToken)
            else:
                listFiles = self.client.list_objects_v2(
                    Bucket=self.s3BucketName, Prefix=path, Delimiter='/')

            # print(listFiles)
            ContinuationToken = listFiles.get('NextContinuationToken', None)
            files = None
            folders = None
            if listFiles is not None:
                files = listFiles.get('Contents')
                folders = listFiles.get('CommonPrefixes')

            if files is not None:
                for file in files:
                    #print("Delete path:%s"%file.get('Key'))
                    self.client.delete_object(
                        Bucket=self.s3BucketName, Key=file.get('Key'))

            if folders is not None:
                for folder in folders:
                    self._s3_removeFolder(folder.get(
                        'Prefix'), remove_self=False)

            if not ContinuationToken:
                break

        if remove_self:
            self.client.delete_object(Bucket=self.s3BucketName, Key=path[:-1])

    def _s3_moveFile(self, oldName, newName):
        self.client.copy_object(Bucket=self.s3BucketName, CopySource={
                                'Bucket': self.s3BucketName, 'Key': oldName}, Key=newName)
        self.client.delete_object(Bucket=self.s3BucketName, Key=oldName)

    def is_file_exists(self, path):
        path = self._get_relative_path(path)
        exists = False
        try:
            res = self.client.head_object(Bucket=self.s3BucketName, Key=path)
            exists = True
        except Exception as e:
            pass

        return exists

    @staticmethod
    def _get_seconds_from_epoch(date_data):
        res = 0
        try:
            if date_data:
                epoch = datetime.datetime(1970, 1, 1, tzinfo=tzutc())
                res = (date_data - epoch).total_seconds()
        except Exception as e:
            #logging.exception("_get_seconds_from_epoch failed.")
            pass

        return res

    def get_mtime(self, path):
        path = self._get_relative_path(path)
        res = None
        try:
            obj = self.client.head_object(Bucket=self.s3BucketName, Key=path)
            res = self._get_seconds_from_epoch(obj.get('LastModified'))
        except Exception as e:
            #logging.exception("getMTime failed.")
            pass

        return res

    def get_file_size(self, path):
        path = self._get_relative_path(path)
        res = 0
        try:
            obj = self.client.head_object(Bucket=self.s3BucketName, Key=path)
            res = obj.get('ContentLength')
        except Exception as e:
            #logging.exception("get_file_size failed.")
            pass

        return res

    def is_folder_exists(self, path):
        path = self._get_relative_path(path)
        listFiles = self.client.list_objects(
            Bucket=self.s3BucketName, Prefix=path + "/", Delimiter='/')
        if listFiles is not None:
            content = listFiles.get('Contents')
            if content is None:
                listFiles = listFiles.get('CommonPrefixes')
            else:
                listFiles = content

        # print(listFiles)
        return listFiles is not None

    def read_text_file(self, path):
        from .local_fsclient import LocalFSClient

        with LocalFSClient().save_atomic(path) as local_tmp_path:
            self.download_file(path, local_tmp_path)
            return LocalFSClient().read_text_file(local_tmp_path)

        # obj = self.client.get_object(Bucket=self.s3BucketName, Key=path)
        # return obj['Body'].read()

    def write_text_file(self, path, data, atomic=False, mode="w"):
        #TODO: support mode="a"
        path = self._get_relative_path(path)
        mimetype, encoding = mimetypes.guess_type(path)
        args = {}
        if mimetype:
            args['ContentType'] = mimetype
        if encoding:
            args['ContentType'] = "application/octet-stream"

        self.client.put_object(Body=data, Bucket=self.s3BucketName, Key=path, **args)

    def copy_file_remote(self, path_src, path_dst):
        path = self._get_relative_path(path_src)
        copy_source = {
            'Bucket': self.s3BucketName,
            'Key': path
        }

        path = self._get_relative_path(path_dst)
        self.client.copy(copy_source, self.s3BucketName, path)

    def copy_file(self, path_src, path_dst):
        if path_src.startswith("s3") and path_dst.startswith("s3"):
            self.copy_file_remote(path_src, path_dst)
        elif path_dst.startswith("s3"):
            path = self._get_relative_path(path_dst)
            self._s3_upload_file(path_src, path)
        elif path_src.startswith("s3"):
            self.download_file(path_src, path_dst)

    def _s3_upload_file(self, path_local, path_s3):
        s3_config = boto3.s3.transfer.TransferConfig(use_threads=False)

        mimetype, encoding = mimetypes.guess_type(path_local)
        args = {}
        if mimetype:
            args['ContentType'] = mimetype
        if encoding:
            args['ContentType'] = "application/octet-stream"

        self.client.upload_file(path_local, Bucket=self.s3BucketName, Key=path_s3, Config=s3_config,
            ExtraArgs=args
        )

    def copy_files(self, path_src, path_dst):
        files = self.list_folder(path_src, wild=True)
        for file in files:
            self.copy_file(os.path.join(os.path.dirname(
                path_src), file), os.path.join(path_dst, file))

    def copy_folder(self, path_src, path_dst):
        from  a2ml.api.utils import fsclient

        files = fsclient.list_folder(path_src)
        for file in files:
            full_src = os.path.join(path_src, file)
            if fsclient.is_file_exists(full_src):
                self.copy_file(full_src, os.path.join(path_dst, file))
            else:
                self.copy_folder(full_src, os.path.join(path_dst, file))

    def download_file(self, path, local_path):
        from .local_fsclient import LocalFSClient

        try:
            from urllib.parse import urlparse
        except ImportError:
            from urlparse import urlparse

        if path.startswith("http:") or path.startswith("https:"):
            s3_path = self._get_relative_path(local_path)

            uri = urlparse(path)

            with LocalFSClient().save_atomic(uri.path) as temp_file:
                LocalFSClient().downloadFile(path, temp_file)
                self._s3_upload_file(temp_file, s3_path)

                #fsclient.waitForFile(local_path, if_wait_for_file=True, num_tries=3000, interval_sec=20)
            # with fsclient.open(path, "rb", encoding=None) as fd:
            #     self.client.upload_fileobj(fd, Bucket=self.s3BucketName, Key=s3_path)
        else:
            path = self._get_relative_path(path)
            LocalFSClient().create_parent_folder(local_path)
            self.client.download_file(
                Bucket=self.s3BucketName, Key=path, Filename=local_path)

    def download_folder(self, path, local_path):
        files = self.list_folder(path)
        for file in files:
            if file.endswith('/'):
                self.download_folder(os.path.join(
                    path, file), os.path.join(local_path, file))
            else:
                self.copy_file(os.path.join(path, file),
                               os.path.join(local_path, file))

    def move_file(self, path_src, path_dst):
        self.copy_file(path_src, path_dst)

        if path_src.startswith("s3"):
            self.remove_file(path_src)
        else:
            from .local_fsclient import  LocalFSClient
            LocalFSClient().remove_file(path_src)



import os
import shutil
import subprocess
import sys
import pandas as pd

from .deploy import ModelDeploy
from ..cloud.cluster import AugerClusterApi
from ..cloud.pipeline import AugerPipelineApi
from ..exceptions import AugerException
from a2ml.api.utils import fsclient, getsizeof_deep
from a2ml.api.utils.dataframe import DataFrame
from a2ml.api.model_review.model_helper import ModelHelper
from ..decorators import with_project
from ..project import Project


class ModelPredict():
    """Predict using deployed Auger Model."""

    def __init__(self, ctx):
        super(ModelPredict, self).__init__()
        self.ctx = ctx

    def execute(self, filename, model_id, threshold=None, locally=False, data=None, columns=None, 
            predicted_at=None, output=None, no_features_in_result=None):
        if filename and not (filename.startswith("http:") or filename.startswith("https:")) and\
            not fsclient.is_s3_path(filename):
            self.ctx.log('Predicting on data in %s' % filename)
            filename = os.path.abspath(filename)

        if locally:
            if locally == "docker":
                predicted = self._predict_locally_in_docker(filename, model_id, threshold, data, columns, predicted_at, output, no_features_in_result)
            else:
                predicted = self._predict_locally(filename, model_id, threshold, data, columns, predicted_at, output, no_features_in_result)
        else:
            predicted = self._predict_on_cloud(filename, model_id, threshold, data, columns, predicted_at, output, no_features_in_result)

        return predicted

    @with_project(autocreate=False)
    def _upload_file_to_cloud(self, project, filename):
        from ..dataset import DataSet

        file_url, file_name = DataSet(self.ctx, project).upload_file(filename)
        return file_url

    def _process_input(self, filename, data, columns):
        send_records = False
        file_url = None
        records = None
        features = None
        max_records_size = 10*1024

        if filename:
            if not (filename.startswith("http:") or filename.startswith("https:")) and \
               not fsclient.is_s3_path(filename):
                file_size = fsclient.get_file_size(filename)
                if file_size <= max_records_size:
                    self.ctx.log("Local file '%s' with size %s(KB) is smaller %s KB, so send records to hub." % (filename, file_size/1024, max_records_size/1024))
                    send_records = True
            else:
                file_url = filename
                if fsclient.is_s3_path(file_url):
                    with fsclient.with_s3_downloaded_or_local_file(file_url) as local_path:
                        file_url = self._upload_file_to_cloud(local_path)
        elif data is not None and isinstance(data, pd.DataFrame):
            if len(data) < 100:
                send_records = True
        else:
            size = getsizeof_deep(data)
            if size is None or size > max_records_size:
                self.ctx.log("Size of data %.2f KB > %.2f KB, so send file to hub." % (size/1024, max_records_size/1024))
            else:
                send_records = True

        if send_records:
            ds = DataFrame.create_dataframe(filename, data, columns)
            records = ds.get_records()
            features = ds.columns
        else:
            if not file_url:
                if not filename:
                    with fsclient.save_atomic("temp_predict.csv.gz", move_file=False) as local_path:
                        ds = DataFrame.create_dataframe(filename, data, columns)
                        ds.saveToCsvFile(local_path)
                        file_url = self._upload_file_to_cloud(local_path)
                else:
                    file_url = self._upload_file_to_cloud(filename)

        return records, features, file_url, data is not None and isinstance(data, pd.DataFrame)

    def _check_model_project(self, pipeline_api):
        model_project_name = Project(self.ctx, project_id=pipeline_api.properties().get('project_id')).properties().get('name')

        if model_project_name != self.ctx.config.get('name'):
            raise AugerException("Project name: %s in config.yml is different from model project name: %s. Please change name in config.yml."%(
                self.ctx.config.get('name'), model_project_name))

    def _predict_on_cloud(self, filename, model_id, threshold, data, columns, predicted_at, output, no_features_in_result):
        records, features, file_url, is_pandas_df = self._process_input(filename, data, columns)
        temp_file = None
        ds_result = None
        if records is not None and len(records) == 0:
            ds_result =  DataFrame.create_dataframe(None, [], features+[self.ctx.config.get('target')])
        else:
            pipeline_api = AugerPipelineApi(self.ctx, None, model_id)
            predictions = pipeline_api.predict(records, features, threshold=threshold, file_url=file_url, 
                predicted_at=predicted_at, no_features_in_result=no_features_in_result)

            try:
                ds_result = DataFrame.create_dataframe(predictions.get('signed_prediction_url'),
                    records=predictions.get('data'), features=predictions.get('columns'))
            except Exception as e:
                self._check_model_project(pipeline_api)

                msg = "Prediction result file(%s) cannot be downloaded."%predictions.get('signed_prediction_url')
                raise AugerException(msg+"Please contact support.")

            temp_file = ds_result.options['data_path'] if predictions.get('signed_prediction_url') else None

        try:
            ds_result.options['data_path'] = filename
            ds_result.loaded_columns = columns
            ds_result.from_pandas = is_pandas_df

            # Save prediction in local model folder if exist
            is_model_loaded, model_path = ModelDeploy(self.ctx, None).verify_local_model(model_id)
            if not is_model_loaded:
                model_path = None

            return ModelHelper.save_prediction(ds_result,
                prediction_id = None, json_result=False, count_in_result=False, prediction_date=predicted_at,
                model_path=model_path, model_id=model_id, output=output)
        finally:
            if temp_file:
                fsclient.remove_file(temp_file)

    def _predict_locally(self, filename_arg, model_id, threshold, data, columns, predicted_at, output, no_features_in_result):
        from auger_ml.model_exporter import ModelExporter

        is_model_loaded, model_path = ModelDeploy(self.ctx, None).verify_local_model(model_id)
        if not is_model_loaded:
            raise AugerException('Model isn\'t loaded locally. '
                'Please use a2ml deploy command to download model.')

        if columns is not None:
            columns = list(columns)

        res, options = ModelExporter({}).predict_by_model_to_ds(model_path, 
            path_to_predict=filename_arg, records=data, features=columns, 
            threshold=threshold, no_features_in_result=no_features_in_result)

        ds_result = DataFrame({'data_path': None})
        ds_result.df = res.df
        ds_result.loaded_columns = columns
        if isinstance(data, pd.DataFrame):
            ds_result.from_pandas = True

        return ModelHelper.save_prediction(ds_result,
            prediction_id = None, json_result=False, count_in_result=False, prediction_date=predicted_at,
            model_path=model_path, model_id=model_id, output=output)

        # return ModelExporter({}).predict_by_model(model_path=model_path,
        #     path_to_predict=filename_arg, records=data, features=columns, 
        #     threshold=threshold, prediction_date=predicted_at, 
        #     no_features_in_result=no_features_in_result) #, output=output)

    def _predict_locally_in_docker(self, filename_arg, model_id, threshold, data, columns, predicted_at, output, no_features_in_result):
        model_deploy = ModelDeploy(self.ctx, None)
        is_model_loaded, model_path = model_deploy.verify_local_model(model_id, add_model_folder=False)
        if not is_model_loaded:
            raise AugerException('Model isn\'t loaded locally. '
                'Please use a2ml deploy command to download model.')

        filename = filename_arg
        if not filename:
            ds = DataFrame.create_dataframe(filename, data, columns)
            filename = os.path.join(self.ctx.config.get_path(), '.augerml', 'predict_data.csv')
            ds.saveToCsvFile(filename, compression=None)

        predicted = self._docker_run_predict(filename, threshold, model_path)

        if not filename_arg:
            ds_result = DataFrame.create_dataframe(predicted)

            ds_result.options['data_path'] = None
            ds_result.loaded_columns = columns

            return ModelHelper.save_prediction(ds_result,
                prediction_id = None,
                json_result=False, count_in_result=False, prediction_date=predicted_at,
                model_path=model_path, model_id=model_id, output=output)
        elif output:
            fsclient.move_file(predicted, output)
            predicted = output

        return predicted

    def _docker_run_predict(self, filename, threshold, model_path):
        cluster_settings = AugerClusterApi.get_cluster_settings(self.ctx)
        docker_tag = cluster_settings.get('kubernetes_stack')
        predict_file = os.path.basename(filename)
        data_path = os.path.abspath(os.path.dirname(filename))
        model_path = os.path.abspath(model_path)

        call_args = "--verbose=True --path_to_predict=./model_data/%s %s" % \
            (predict_file, "--threshold=%s" % str(threshold) if threshold else '')

        command = (r"docker run "
            "-v {model_path}:/var/src/auger-ml-worker/exported_model "
            "-v {data_path}:/var/src/auger-ml-worker/model_data "
            "deeplearninc/auger-ml-worker:{docker_tag} "
            "python ./exported_model/client.py {call_args}").format(
                model_path=model_path, data_path=data_path,
                docker_tag=docker_tag, call_args=call_args)

        try:
            self.ctx.log(
                'Running model in deeplearninc/'
                'auger-ml-worker:%s' % docker_tag)
            result_file = subprocess.check_output(command, stderr=subprocess.STDOUT, shell=True)
            result_file = result_file.decode("utf-8").strip()
            result_file = os.path.basename(result_file)
            # getattr(subprocess,
            #     'check_call' if self.ctx.debug else 'check_output')(
            #     command, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as e:
            raise AugerException('Error running Docker container...')

        return os.path.join(data_path, "predictions", result_file)



import arff
import copy
import json
import logging
import math
import os
import pandas as pd
import warnings

from functools import wraps

from a2ml.api.utils import fsclient, get_uid, get_uid4, remove_dups_from_list, process_arff_line, download_file, retry_helper, parse_url
from a2ml.api.utils.local_fsclient import LocalFSClient


# To avoid warnings for inplace operation on datasets
pd.options.mode.chained_assignment = None

class DataFrame(object):
    BOOLEAN_WORDS_TRUE = ['yes', 'on']
    BOOLEAN_WORDS_FALSE = ['no', 'off']

    def __init__(self, options):
        self.options = options
        self.categoricals = {}
        self.transforms_log = [[],[],[],[]]
        self.df = None
        self.dataset_name = None
        self.loaded_columns = None
        self.from_pandas = False

    def _get_compression(self, extension):
        compression = self.options.get('data_compression', 'infer')
        if extension.endswith('.gz') or extension.endswith('.gzip'):
            compression = 'gzip'
        elif extension.endswith('.bz2'):
            compression = 'bz2'
        elif extension.endswith('.zip'):
            compression = 'zip'
        elif extension.endswith('.xz'):
            compression = 'xz'

        return compression

    @staticmethod
    def create_dataframe(data_path=None, records=None, features=None):
        if data_path:
            ds = DataFrame({'data_path': data_path})
            ds.load(features = features)
        elif records is not None and isinstance(records, pd.DataFrame):
            ds = DataFrame({})
            ds.df = records
            ds.from_pandas = True
        else:
            ds = DataFrame({})
            ds.load_records(records, features=features)

        return ds

    @staticmethod
    def load_from_files(files, features=None):
        for file in files:
            path = file if type(file) == str else file['path']

            fsclient.wait_for_file(path, True)
            try:
                df = retry_helper(lambda: DataFrame.create_dataframe(path, None, features))
                yield (file, df)
            except Exception as exc:
                logging.exception("load_from_files failed for: %s. Error: %s"%(path, exc))

    def load_from_file(self, path, features=None, nrows=None):
        from collections import OrderedDict

        extension = path
        if self.options.get('data_extension', 'infer') != 'infer':
            extension = self.options['data_extension']

        if self.options.get('content_type') == 'multipart':
            fsclient.merge_folder_files(path)

        if extension.endswith('.arff') or extension.endswith('.arff.gz'):
            arffFile = None
            class ArffFile:
                def __init__(self, file):
                    self.file = file
                    self.date_attrs = {}

                def __iter__(self):
                    return self

                def __next__(self):
                    line = process_arff_line(next(self.file), self.date_attrs)
                    return line

            try:

                with fsclient.open(path, 'r') as f:
                    arffFile = ArffFile(f)
                    arff_data = arff.load(arffFile, return_type=arff.COO)

                convert_arff = DataFrame._convert_arff_coo
            except arff.BadLayout:
                with fsclient.open(path, 'r') as f:
                    arffFile = ArffFile(f)
                    arff_data = arff.load(arffFile, return_type=arff.DENSE)

                convert_arff = DataFrame._convert_arff_dense

            columns = [a[0] for a in arff_data['attributes']]
            series = convert_arff(features, columns, arff_data['data'])

            res = pd.DataFrame.from_dict(OrderedDict(
                (c, s) for c, s in zip(columns, series) if s is not None
            ))
            for date_field, fmt in arffFile.date_attrs.items():
                res[date_field] = pd.to_datetime(res[date_field], infer_datetime_format=True, errors='ignore', utc=True)

            return res
        elif extension.endswith('.pkl') or extension.endswith('.pkl.gz'):
            return self.loadFromBinFile(path, features)
        elif extension.endswith('.json') or extension.endswith('.json.gz'):
            path = fsclient.s3fs_open(path)
            res = pd.read_json(path, orient=self.options.get('json_orient',None))
            if features:
                res = res[features]
            return res
        elif extension.endswith('.xlsx') or extension.endswith('.xls'):
            path = fsclient.s3fs_open(path)
            return pd.read_excel(path, usecols=features)
        elif extension.endswith('.feather') or extension.endswith('.feather.gz') or extension.endswith('.feather.zstd') or extension.endswith('.feather.lz4'):
            # Features list is optional for feather file, but it can segfault without it on some files
            return self.loadFromFeatherFile(path, features)
        elif extension.endswith('.parquet'):
            return self.loadFromParquetFile(path, features)

        csv_with_header = self.options.get('csv_with_header', True)
        header = 0 if csv_with_header else None
        prefix = None if csv_with_header else 'c'

        compression = self._get_compression(extension)
        path = fsclient.s3fs_open(path)

        res_df = None
        try:
            res_df = pd.read_csv(
                path,
                encoding='utf-8',
                escapechar="\\",
                usecols=features,
                na_values=['?'],
                header=header,
                prefix=prefix,
                sep = ',',
                nrows=nrows,
                low_memory=False,
                compression=compression
            )
        except Exception as e:
            logging.error("read_csv failed: %s"%e)
            res_df = pd.read_csv(
                path,
                encoding='utf-8',
                escapechar="\\",
                usecols=features,
                na_values=['?'],
                header=header,
                prefix=prefix,
                sep = '|',
                nrows=nrows,
                low_memory=False,
                compression=compression
            )

        # if res_df is not None:
        #     for name, value in res_df.dtypes.items():
        #         if value == 'object':
        #             res_df[name] = pd.to_datetime(res_df[name], infer_datetime_format=True, errors='ignore', utc=True)

        return res_df

    def load(self, features=None, nrows=None):
        self.categoricals = {}
        self.transforms_log = [[],[],[],[]]

        import csv
        from io import StringIO

        path = self.options['data_path']
        if isinstance(path, StringIO):
            path.seek(0)
            self.df = pd.read_csv(path, encoding='utf-8', escapechar="\\", usecols=features, na_values=['?'], nrows=nrows)
            if self.options.get("targetFeature") in self.df.columns:
                self.dropna([self.options["targetFeature"]])
        else:
            if path.startswith("jdbc:"):
                import psycopg2
                from psycopg2.extensions import parse_dsn
                path = path.replace('sslfactory=org.postgresql.ssl.NonValidatingFactory&', '')
                path, params = parse_url(path)

                self.dbconn_args = parse_dsn(path)
                conn = psycopg2.connect(**self.dbconn_args)

                sql_cmd = "select " + (",".join(features) if features else "*") +" from %s"%params['tablename'][0]
                if 'limit' in params:
                    sql_cmd += " LIMIT %s"%params['limit'][0]

                if 'offset' in params:
                    sql_cmd += " OFFSET %s"%params['offset'][0]

                logging.info("Read data from remote DB: %s"%sql_cmd)
                self.df = pd.read_sql(sql_cmd, con=conn)
            else:
                path, remote_path = self._check_remote_path()
                try:
                    self.df = self.load_from_file(path, features=features, nrows=nrows)
                except:
                    if remote_path:
                        logging.exception("Loading local file failed. Download it again...")
                        self.options['data_path'] = remote_path
                        path, remote_path = self._check_remote_path(force_download=True)
                        self.df = self.load_from_file(path, features=features, nrows=nrows)
                    else:
                        raise

                self.dataset_name = os.path.basename(path)

            if self.options.get("targetFeature") in self.df.columns:
                self.dropna([self.options["targetFeature"]])
        return self

    def _check_remote_path(self, force_download=False):
        remote_path = None
        if self.options['data_path'].startswith("http:") or self.options['data_path'].startswith("https:"):
            local_dir = LocalFSClient().get_temp_folder()
            file_name = 'data-' + get_uid4()

            local_file_path = download_file(self.options['data_path'],
                local_dir=local_dir, file_name=file_name, force_download=force_download)

            remote_path = self.options['data_path']
            self.options['data_path'] = local_file_path

        return self.options['data_path'], remote_path

    def load_records(self, records, features=None):
        self.categoricals = {}
        self.transforms_log = [[],[],[],[]]

        if features:
            self.df = pd.DataFrame.from_records(records, columns=features)
            self.loaded_columns = features
        else:
            self.df = pd.DataFrame(records) #dict

        return self

    def get_records(self):
        return self.df.values.tolist()

    def saveToCsvFile(self, path, compression="gzip"):
        fsclient.remove_file(path)
        fsclient.create_parent_folder(path)

        with fsclient.save_local(path) as local_path:
            self.df.to_csv(local_path, index=False, compression=compression, encoding='utf-8')

    def saveToBinFile(self, path):
        fsclient.save_object_to_file(self.df, path)

    def loadFromBinFile(self, path, features=None):
        self.df = fsclient.load_object_from_file(path)

        if features:
            self.df =  self.df[features]

        return self.df

    def saveToFeatherFile(self, path):
        fsclient.save_object_to_file(self.df, path, fmt="feather")

    def saveToParquetFile(self, path, compression="gzip"):
        fsclient.remove_file(path)
        fsclient.create_parent_folder(path)

        with fsclient.save_local(path) as local_path:
            self.df.to_parquet(local_path, index=False, compression=compression)

    def loadFromFeatherFile(self, path, features=None):
        self.df = fsclient.load_db_from_feather_file(path, features)
        return self.df

    def loadFromParquetFile(self, path, features=None):
        self.df = fsclient.load_db_from_parquet_file(path, features)
        return self.df

    def saveToFile(self, path):
        if path.endswith('.feather') or path.endswith('.feather.gz') or path.endswith('.feather.zstd') or path.endswith('.feather.lz4'):
            self.saveToFeatherFile(path)
        elif path.endswith('.parquet'):
            self.saveToParquetFile(path)
        else:
            compression = None
            if path.endswith('.gz'):
                compression = 'gzip'

            self.saveToCsvFile(path, compression)

    def count(self):
        if self.df is not None:
            return len(self.df)
        else:
            return 0

    def __len__(self):
        return self.count()

    @property
    def columns(self):
        return self.df.columns.tolist()

    def _map_dtypes(self, dtype):
        dtype_map = {'int64': 'integer', 'float64':'double', 'object': 'string',
            'categorical':'categorical', 'datetime64[ns]': 'datetime', 'bool': 'boolean'}
        if dtype_map.get(dtype, None):
            return dtype_map[dtype]

        if dtype and (dtype.startswith('int') or dtype.startswith('uint')):
            return 'integer'

        if dtype and dtype.startswith('float'):
            return 'float'

        if dtype and dtype.startswith('double'):
            return 'double'

        if dtype and dtype.startswith('datetime64'):
            return 'datetime'

        return dtype

    @property
    def dtypes(self):
        types_list = []
        columns_list = self.columns
        for idx, dtype in enumerate(self.df.dtypes):
            types_list.append((columns_list[idx], self._map_dtypes(dtype.name)))

        return types_list

    @property
    def dtypes_dict(self):
        types_dict = {}
        columns_list = self.columns
        for idx, dtype in enumerate(self.df.dtypes):
            types_dict[columns_list[idx]] = self._map_dtypes(dtype.name)

        return types_dict

    def select(self, features):
        self.df = self.df[features]
        return self

    def drop(self, columns):
        self.df.drop(columns, inplace=True, axis=1)

    def drop_duplicates(self, columns=None):
        self.df.drop_duplicates(subset=columns, inplace=True)
        self.df.reset_index(drop=True, inplace=True)
        return self

    def dropna(self, columns=None):
        self.df.dropna(subset=columns, inplace=True, axis=0)
        self.df.reset_index(drop=True, inplace=True)
        return self

    def fillna(self, value):
        if isinstance(value, dict):
            value = value.copy()
            for item in self.dtypes:
                if list(value.keys())[0] == item[0]:
                    if item[1] == 'string':
                        value[list(value.keys())[0]] = str(list(value.values())[0])
                    elif item[1] == 'integer':
                        value[list(value.keys())[0]] = int(list(value.values())[0])
                    else:
                        value[list(value.keys())[0]] = float(list(value.values())[0])

        self.df.fillna(value, inplace=True)
        return self

    def convertToCategorical(self, col_names, is_target = False, categories = None):
        #print("convertToCategorical:%s"%col_names)
        if not isinstance(col_names, list):
            col_names = [col_names]

        if is_target:
            for col in col_names:
                if col in self.columns and self.categoricals.get(col, None) is None:
                    self.df[col] = pd.Categorical(self.df[col], categories=categories)
                    self.categoricals[col] = {'categories': list(self.df[col].cat.categories)}
                    self.df[col] = self.df[col].cat.codes
        else:
            cols_to_process = []
            cols = self.columns
            for col in col_names:
                if col in cols:
                    cols_to_process.append(col)

            #print(cols_to_process)
            if cols_to_process:
                self.df = pd.get_dummies(self.df, columns=cols_to_process)
                new_cols = self.columns

                for col in cols_to_process:
                    generated_cols = []
                    for new_col in new_cols:
                        if new_col.startswith(col+'_'):
                            generated_cols.append(new_col)

                    self.categoricals[col] = {'columns': generated_cols}

        return self

    @staticmethod
    def _convert_arff_coo(features, columns, arff_data_data):
        if features is None:
            data = [([], []) for _ in columns]
        else:
            fset = remove_dups_from_list(features)
            data = [([], []) if c in fset else None for c in columns]

        for v, i, j in zip(*arff_data_data):
            d = data[j]
            if d is not None:
                indices, values = d
                if indices:
                    assert indices[-1] < i
                indices.append(i)
                values.append(v)

        max_i = -1
        for d in data:
            if d is not None and len(d[0]) > 0:
                max_i = max(max_i, d[0][-1])
        height = max_i + 1

        series = []
        for d in data:
            if d is None:
                s = None
            else:
                keys, values = d
                sa = pd.SparseArray(
                    values,
                    sparse_index=pd._libs.sparse.IntIndex(height, keys),
                    fill_value=0
                )
                s = pd.Series(sa.values)
            series.append(s)

        return series

    @staticmethod
    def _convert_arff_dense(features, columns, arff_data_data):
        if features is None or set(features) == set(columns):
            return zip(*arff_data_data)

        fset = remove_dups_from_list(features)
        return [
            [row[i] for row in arff_data_data] if c in fset else None
            for i, c in enumerate(columns)
        ]

    def _check_for_json(self, cname):
        rows = self.df[cname][(self.df[cname] != '[]') & (self.df[cname] != '{}')].values

        if rows is None or len(rows) == 0:
            return []
        else:
            row = rows[0]

        try:
            data = json.loads(row)
            children = []
            for key in data:
                name = str(key) + '_' + cname
                dtype = type(data[key]).__name__
                if dtype == 'unicode':
                    dtype = 'string'

                children.append(
                    {
                        'column_name': name,
                        'orig_column_name': str(key),
                        'child': cname,
                        'parent': None,
                        'orig_datatype': dtype,
                        'datatype': dtype,
                        'range': ('', ''),
                        'use': False,
                        'isTarget': False,
                        'missing_values': 0,
                        'unique_values': 0
                    }
                )
            return children
        except Exception as e:
            return []

    def __format_number_for_serialization(self, v):
        if v != v or math.isinf(v):
            return None
        else:
            return round(v, 6)

    def getSummary(self):
        types_list = []
        columns_list = self.df.columns.values.tolist()
        for idx, dtype in enumerate(self.df.dtypes):
            types_list.append((columns_list[idx], self._map_dtypes(dtype.name)))

        info = {"dtypes": self.dtypes, "count": self.count(), 'columns_count': len(self.columns)}
        stat_data = []
        for x in info['dtypes']:
            cname = str(x[0])
            ctype = x[1]
            children = []

            if ctype == 'string':
                children = self._check_for_json(cname)

            # add children or just add row
            if children:
                stat_data = stat_data + children
            else:
                item = {
                    'column_name': cname,
                    'orig_datatype': ctype,
                    'datatype': ctype,
                }

                if ctype == 'integer' or ctype == 'float' or ctype == 'double':
                    mean = self.__format_number_for_serialization(self.df[cname].mean())
                    if mean:
                        item['avg'] = mean

                    std = self.__format_number_for_serialization(self.df[cname].std())
                    if std:
                        item['std_dev'] = std

                stat_data.append(item)

        #TODO : remove from UI isTarget and use
        lef = set(self.options.get('labelEncodingFeatures', []))
        tf = self.options.get('targetFeature', "")
        fc = set(self.options.get('featureColumns', []))
        tsf = set(self.options.get('timeSeriesFeatures', []))
        dtf = self.options.get('datetimeFeatures', [])
        if dtf is None:
            dtf = []

        for row in stat_data:
            row['isTarget'] = False
            row['use'] = False

            cname = row['column_name']
            if row['datatype'] == 'string':
                row['datatype'] = 'hashing' if cname in lef else 'categorical'

            if cname == tf:
                row['isTarget'] = True
            if cname in fc:
                row['use'] = True
            if cname in tsf:
                row['datatype'] = 'timeseries'
            if cname in dtf:
                row['datatype'] = 'datetime'

        info['stat_data'] = stat_data
        del info['dtypes']

        return info

    def update_options_by_dataset_statistics(self, stat_data=None):
        transforms = []
        categoricals = []
        hashings = []
        timeseries = []
        selected_features = []
        target_feature = None
        binaryClassification = None
        minority_target_class = None
        datetimeFeatures = []
        stat_data = stat_data or self.options.get('dataset_statistics', {}).get('stat_data', [])

        for item in stat_data:
            if item.get('isTarget'):
                target_feature = item['column_name']
                if (self.options.get('model_type', '') == 'classification' or self.options.get('classification')):
                    binaryClassification = item.get('unique_values', 0) == 2

                    #Find minority target class
                    if item.get('value_counts_ex'):
                        minority_target_class = item['value_counts_ex'][0]['value']
                        minority_target_count = item['value_counts_ex'][0]['count']
                        for vc in item['value_counts_ex']:
                            if vc['count'] < minority_target_count:
                                minority_target_class = vc['value']
                                minority_target_count = vc['count']

            if item.get('use') and not item.get('isTarget'):
                selected_features.append(item['column_name'])

            if item.get('use') or item.get('isTarget'):
                if item['orig_datatype'] == 'string':
                    if item['datatype'] == 'integer' or item['datatype'] == 'double':
                        transforms.append({"withNumericColumn":{"col_name":item['column_name']}})
                    if item['datatype'] == 'boolean':
                        transforms.append({"withBooleanColumn":{"col_name":item['column_name']}})

                if item['datatype'] == 'categorical':
                    categoricals.append(item['column_name'])
                if item['datatype'] == 'hashing':
                    categoricals.append(item['column_name'])
                    hashings.append(item['column_name'])
                if item['datatype'] == 'timeseries':
                    timeseries.append(item['column_name'])
                if item['datatype'] == 'datetime':
                    datetimeFeatures.append(item['column_name'])

        if categoricals:
            self.options['categoricalFeatures'] = categoricals
        if hashings:
            self.options['labelEncodingFeatures'] = hashings
        if timeseries:
            self.options['timeSeriesFeatures'] = timeseries
        if selected_features:
            self.options['featureColumns'] = selected_features
        if target_feature:
            self.options['targetFeature'] = target_feature
        if datetimeFeatures:
            self.options['datetimeFeatures'] = datetimeFeatures
        if binaryClassification is not None:
            self.options['binaryClassification'] = binaryClassification
        if minority_target_class is not None and self.options.get('minority_target_class') is None and \
            self.options.get('use_minority_target_class'):
            self.options['minority_target_class'] = minority_target_class

        if not 'datasource_transforms' in self.options:
            self.options['datasource_transforms'] = []

        self.options['datasource_transforms'] = [transforms]

        return self.options



import os
import csv

from ..cloud.pipeline import AugerPipelineApi
from .predict import ModelPredict
from a2ml.api.utils import fsclient

class ModelActual():
    """Predict using deployed Auger Model."""

    def __init__(self, ctx):
        super(ModelActual, self).__init__()
        self.ctx = ctx

    def execute(self, model_id, filename, data, columns, actuals_at, actual_date_column):
        records, features, file_url, is_pandas_df = ModelPredict(self.ctx)._process_input(filename, data, 
            columns=columns)
        pipeline_api = AugerPipelineApi(self.ctx, None, model_id)

        return pipeline_api.actual(records, features, actuals_at, file_url, actual_date_column)




from a2ml.api.utils.provider_runner import ProviderRunner

class BaseA2ML(object):
    def __init__(self, ctx, object_name):
        self.ctx = ctx
        self.object_name = object_name

    def build_runner(self, ctx, provider, force_local=False):
        if not ctx.use_auger_cloud() and ctx.config.get('use_server', False) and not force_local:
            from a2ml.api.utils.remote_runner import RemoteRunner

            return RemoteRunner(ctx, provider, self.object_name)
        elif self.object_name:
            from a2ml.api.utils.crud_runner import CRUDRunner

            return CRUDRunner(ctx, provider, self.object_name)
        else:
            return ProviderRunner(ctx, provider)

    def get_runner(self, locally, model_id=None, provider=None):
        if provider is None and model_id and not self.ctx.is_external_provider():
            provider = self.ctx.get_model_provider(model_id)

        if provider:
            return self.build_runner(self.ctx, provider, force_local=locally)
        else:
          if locally:
              return self.local_runner()
          else:
              return self.runner



import json
import numpy as np

from operator import attrgetter
from itertools import groupby

from a2ml.api.roi.base_interpreter import BaseInterpreter
from a2ml.api.roi.lexer import AstError, Token
from a2ml.api.roi.validator import Validator

class InterpreterError(AstError):
    pass

class MissedVariable(InterpreterError):
    pass

class TopRecord:
    def __init__(self, row, group_key, order_value):
        self.row = row
        self.group_key = group_key
        self.order_value = order_value

class Interpreter(BaseInterpreter):
    def __init__(self, expression, vars_mapping={}):
        self.expression = expression
        self.vars_mapping = vars_mapping

    def run(self, variables={}, filter=False):
        known_vars = self.get_known_vars(variables) | set(self.vars_mapping.keys())
        validator = Validator(self.expression, known_vars)
        validation_result = validator.validate(force_raise=True)
        self.root = validation_result.tree

        if isinstance(variables, list):
            return self.evaluate_for_list(self.root, variables, filter=filter)
        else:
            self.variables = variables
            return self.evaluate(self.root)

    def evaluate_for_list(self, node, rows, filter=False):
        # top expressions requires aggregation
        if node.require_aggregation:
            return self.evaluate(node, rows)
        else:
            res = []

            for vars in rows:
                self.variables = vars
                value = self.evaluate(node)
                if filter:
                    if value:
                        res.append(vars)
                else:
                    res.append(value)

            return res

    '''
    Extract names from variable list
    if variables is a dict just return all keys
    if variables is a list of dicts extract keys from each line and return their intersection
    '''
    def get_known_vars(self, variables):
        if isinstance(variables, list):
            if len(variables) > 0:
                res = set(variables[0])

                for vars in variables:
                    res = res.intersection(set(vars.keys()))

                return res
            else:
                return set()
        else:
            return set(variables.keys())

    def evaluate_no_op_node(self, node):
        pass

    def evaluate_const_node(self, node):
        return node.value

    def evaluate_var_node(self, node):
        var_name = self.vars_mapping.get(node.name, node.name)
        if var_name in self.variables:
            return self.variables[var_name]
        elif var_name.startswith("$") and var_name[1:] in self.variables:
            # for non-known vars try to just look up in row
            return self.variables[var_name[1:]]
        else:
            raise MissedVariable(f"missed var `{var_name}` in row `{json.dumps(self.variables)}`")


    def evaluate_binary_op_node(self, node):
        left = self.evaluate(node.left)
        right = self.evaluate(node.right)

        if node.op == Token.MUL:
            return left * right
        elif node.op == Token.DIV:
            return left / right
        elif node.op == Token.INT_DIV:
            return left // right
        elif node.op == Token.MODULO:
            return left % right
        elif node.op == Token.PLUS:
            return left + right
        elif node.op == Token.MINUS:
            return left - right
        elif node.op == Token.POWER:
            return left ** right
        elif node.op == Token.GT:
            return left > right
        elif node.op == Token.GTE:
            return left >= right
        elif node.op == Token.LT:
            return left < right
        elif node.op == Token.LTE:
            return left <= right
        elif node.op == Token.EQ2 or node.op == Token.EQ:
            return left == right
        elif node.op == Token.NE:
            return left != right
        elif node.op == Token.BIT_XOR:
            return left ^ right
        elif node.op == Token.BIT_OR:
            return left | right
        elif node.op == Token.BIT_AND:
            return left & right
        elif node.op == Token.BIT_LSHIFT:
            return left << right
        elif node.op == Token.BIT_RSHIFT:
            return left >> right
        elif node.op == Token.IN:
            return left in right
        elif node.op == Token.OR:
            return left or right
        elif node.op == Token.AND:
            return left and right
        else:
            raise self.error(f"unknown binary operator '{node.op}'")

    def evaluate_unary_op_node(self, node):
        value = self.evaluate(node.node)

        if node.op == Token.PLUS:
            return value
        elif node.op == Token.MINUS:
            return -value
        elif node.op == Token.BIT_NOT:
            return ~value
        elif node.op == Token.NOT:
            return not value
        else:
            raise self.error(f"unknown unary operator '{node.op}'")

    def evaluate_func_node(self, node, rows=None):
        func = self.func_values()[node.func_name]

        if Interpreter.is_static_func(func):
            func_args = [self]
        else:
            func_args = []

        if node.func_name in ("if", "@if"):
            func_args += node.arg_nodes
        elif rows:
            func_args += list(map(lambda n: self.evaluate_for_list(n, rows), node.arg_nodes))
        else:
            func_args += list(map(lambda n: self.evaluate(n), node.arg_nodes))

        return func(*func_args)

    def evaluate_top_node(self, node, rows):
        if node.nested_node:
            rows = self.evaluate(node.nested_node, rows)

        if node.where_node:
            rows = self.evaluate_for_list(node.where_node, rows, filter=True)

        if node.group_node:
            group_keys = self.evaluate_for_list(node.group_node, rows)
        else:
            group_keys = np.zeros(len(rows))

        if node.order_node:
            order_values = self.evaluate_for_list(node.order_node, rows)
        else:
            order_values = np.arange(len(rows))

        table = [TopRecord(row, group_keys[i], order_values[i]) for i, row in enumerate(rows)]
        table.sort(key=attrgetter('group_key'))

        if node.limit_node:
            limit = self.evaluate(node.limit_node)
        else:
            limit = np.inf

        result = []

        for key, records in groupby(table, attrgetter('group_key')):
            records = list(records)

            if node.with_node:
                for with_item_node in node.with_node.with_item_nodes:
                    col_value = self.evaluate(with_item_node.source_node, [item.row for item in records])
                    col_name = with_item_node.alias()

                    for item in records:
                        item.row[col_name] = col_value

            if node.having_node:
                # Drop whole group if none of the rows meet having expression
                having = self.evaluate_for_list(node.having_node, [item.row for item in records])
                if not any(having):
                    records = []

            records.sort(reverse=node.kind == Token.TOP, key=attrgetter('order_value'))
            for ri in range(min(len(records), limit)):
                result.append(records[ri])

        result.sort(key=attrgetter('order_value'), reverse=node.kind == Token.TOP)
        return [item.row for item in result]

    def error(self, msg):
        raise InterpreterError(msg)



import time
import pytest
from a2ml.cmdl.cmdl import cmdl


class TestA2mlCLI():
    def test_import_train_evaluate_deploy_predict(self, runner, log, project):
        result = runner.invoke(cmdl, ['project', 'delete', 'cli-integration-test'])

        result = runner.invoke(cmdl, ['import'])
        assert log.messages[-2] == '[auger]  Created DataSet iris.csv on Auger Cloud.'
        assert log.messages[-1] == '[auger]  DataSet name stored in auger.yaml/dataset'

        result = runner.invoke(cmdl, ['dataset', 'list'])
        assert log.messages[-2] == '[auger]  [x] iris.csv'
        assert log.messages[-1] == '[auger]  1 DataSet(s) listed'

        result = runner.invoke(cmdl, ['train'])
        assert log.messages[-2] == '[auger]  Created Experiment iris.csv-experiment '
        assert log.messages[-1] == '[auger]  Started Experiment iris.csv-experiment training.'

        while True:
            result = runner.invoke(cmdl, ['evaluate'])
            if log.messages[-1] in [
                '[auger]  Search is preprocessing data for training...',
                '[auger]  Search is in progress...']:
                time.sleep(10)
                continue
            elif log.messages[-1] == '[auger]  Search is completed.':
                model_id = log.messages[-3]\
                    .replace(' ', '').replace('[auger]','').split('|')[0]
                break
            else:
                assert log.messages[-1] == 'unknown message'

        print('model is: %s' % model_id)

        result = runner.invoke(cmdl, ['deploy', model_id, '--locally'])
        assert log.messages[-1] == '[auger]  Pulling docker image required to predict'

        result = runner.invoke(cmdl, ['predict', 'iris.csv', '-m', model_id, '--locally'])
        assert log.messages[-1].split('/')[-1] == 'iris_predicted.csv'

        result = runner.invoke(cmdl, ['project', 'delete', 'cli-integration-test'])



import os
import sys
import click
import logging
from .config import Config
import warnings

log = logging.getLogger("a2ml")

CONTEXT_SETTINGS = dict(auto_envvar_prefix='A2ML')
PROVIDERS = ['auger', 'google', 'azure', 'external']
PROVIDERS_META = '|'.join(PROVIDERS)


class Context(object):
    """The Context class provides an environment to run A2ML"""
    def __init__(self, name='config', path=None, debug=False):
        """Initializes the Context instance
        
        Args:
            name (str): The name of the config file. Default is 'config'
            path(str): The path to your config file. If the config file is in the root directory leave as None.
            debug (bool): True | False. Default is False.
        
        Returns:
            object: Context object

        Example:
            .. code-block:: python

                ctx = Context()
        """
        super(Context, self).__init__()

        self.config = Config(name=name, path=path)
        self.name = self.config.name
        self.notificator = None
        self.request_id = None
        self.provider_info = None

        if len(self.name) > 0:
            self.name = "{:<9}".format('[%s]' % self.name)
        self.debug = self.config.get('debug', debug)
        self.set_runs_on_server(False)

        def warn(*args, **kwargs):
            self.system_warning(*args, **kwargs)

        warnings.warn = warn

    def set_runs_on_server(self, value):
        self._runs_on_server = value
        self.config.runs_on_server = value

    def is_runs_on_server(self):
        return self._runs_on_server
            
    def use_auger_cloud(self):
        return self.config.get('use_auger_cloud', self.config.get('use_a2ml_hub', False))

    def get_name(self):
        return self.config.name

    def get_providers(self, provider = None):
        """constructs Context instance

        Args:
            name (str): The name of the config file. Default is 'config'
            path(str): The path to your config file. If the config file is in the root directory leave as None.
            debug (bool): True | False. Default is False.

        Returns:
            list[str]: ['azure', 'auger']

        Examples:
            .. code-block:: python

                ctx = Context()
                ctx.get_providers()

        """
        if provider:
            providers = provider
        else:
            providers = self.config.get('providers', [])

        if isinstance(providers, (str,)):
            providers = [p.strip() for p in providers.split(',')]

        if isinstance(providers, (list,)):
            for p in providers:
                if p not in PROVIDERS:
                    raise Exception('Provider %s is not supported.' % p)
            return providers

        raise Exception('Expecting list of providers in config.yaml\providers')

    def get_model_provider(self, model_id):
        if not model_id:
            return None
            
        if model_id.startswith("AutoML_"):
            return "azure"

        return "auger"

    def is_external_provider(self):
        return self.config.get_list('providers')[0] == 'external'

    def copy(self, name):
        """creates a copy of an existing Context
        
        Args:
            name (str): The name of the config file. Default is 'config'

        Returns:
            object: Context object

        Example:
            .. code-block:: python

                ctx = Context()
                new_ctx = ctx.copy()
        """
        new = Context(name, self.config.path, self.debug)
        new.set_runs_on_server(self._runs_on_server)
        new.notificator = self.notificator
        new.request_id = self.request_id
        new.config.parts = self.config.parts
        new.config.parts_changes = self.config.parts_changes

        try:
            new.config.set("providers", name, config_name='config')
        except Exception as e:
            # In case if command run in folder without config, do not set it
            pass
            
        if self._runs_on_server and hasattr(self, 'credentials'):
            new.credentials = self.credentials

        return new

    def log(self, msg, *args, **kwargs):
        log.info('%s%s' %(self.name, msg), *args, **kwargs)
        self.publish_log('info', '%s%s' %(self.name, msg), *args, **kwargs)

    def log_debug(self, msg, *args, **kwargs):
        if self.debug:
            log.debug('%s%s' %(self.name, msg), *args, **kwargs)
            self.publish_log('debug', '%s%s' %(self.name, msg), *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        log.error('%s%s' %(self.name, msg), *args, **kwargs)
        self.publish_log('error', '%s%s' %(self.name, msg), *args, **kwargs)

    def system_warning(self, *args, **kwargs):
        if self.debug and args:
            name = '[warning] '
            msg = args[0]
            log.warning('%s%s' %(name, msg))
            self.publish_log('warning', '%s%s' %(name, msg))

    def publish_log(self, level, msg, *args, **kwargs):
        if self.notificator:
            self.notificator.publish_log(self.request_id, level, msg, args, kwargs)


    @staticmethod
    def setup_logger(format='%(asctime)s %(name)s | %(message)s'):
        logging.basicConfig(
            stream=sys.stdout,
            datefmt='%H:%M:%S',
            format=format,
            level=logging.INFO)


pass_context = click.make_pass_decorator(Context, ensure=True)




import os

from  a2ml.api.utils import fsclient


class Template(object):

    @staticmethod
    def copy_config_files(experiment_path, providers):
        module_path = os.path.dirname(os.path.abspath(__file__))
        for provider in providers:
            src_config = os.path.abspath(os.path.join(
                module_path, '../template/%s.yaml' % provider))
            dest_config = os.path.join(experiment_path, '%s.yaml' % provider)

            fsclient.copy_file(src_config, dest_config)



# Lexical Grammar:
#
# NUMBER         → DIGIT+ ( "." DIGIT+ )? ;
# STRING         → "\"" <any char except "\"">* "\"" ;
# IDENTIFIER     → (ALPHA | "@") ( ALPHA | DIGIT )* ;
# ALPHA          → "a" ... "z" | "A" ... "Z" | "_" ;
# DIGIT          → "0" ... "9" ;

class AstError(Exception):
    def __init__(self, msg, position=None):
        self.position = position
        super().__init__(msg)

class LexerError(AstError):
    pass

class Token:
    EQ = "="
    EQ2 = "=="
    GT = ">"
    LT = "<"
    NE = "!="
    GTE = ">="
    LTE = "<="

    PLUS = "+"
    MINUS = "-"
    MUL = "*"
    POWER = "**"
    DIV = "/"
    INT_DIV = "//"
    MODULO = "%"

    LPAREN = "("
    RPAREN = ")"
    BIT_OR = "|"
    BIT_XOR = "^"
    BIT_AND = "&"
    BIT_NOT = "~"
    BIT_LSHIFT = "<<"
    BIT_RSHIFT = ">>"

    EXCLAMATION = "!"
    SEMI = ";"
    DOT = "."
    AT = "@"
    UNDER = "_"
    LCURCLY = "{"
    RCURCLY = "}"
    COLON = ":"
    ASSIGN = ":="
    COMMA = ","
    DQUOTE = '"'
    DOLLAR = "$"

    INT_CONST = "INT_CONST"
    FLOAT_CONST = "FLOAT_CONST"
    STR_CONST = "STR_CONST"
    CONST = "CONST"
    ID = "ID"
    AND = "and"
    OR = "or"
    NOT = "not"
    IN = "in"

    ALL = "all"
    AS = "as"
    BOTTOM = "bottom"
    BY = "by"
    FROM = "from"
    HAVING = "having"
    PER = "per"
    TOP = "top"
    WHERE = "where"
    WITH = "with"

    EOF = "EOF"

    COMPARISON_SYMBOLS = set([LT, EQ, GT, EXCLAMATION])
    KEYWORDS = set([AND, OR, NOT, IN, TOP, BOTTOM, BY, PER, WHERE, FROM, HAVING, AS, WITH, ALL])

    SYMBOLS = set(
        [SEMI, DOT, PLUS, MINUS, MUL, DIV, LPAREN, RPAREN, COMMA, BIT_OR, BIT_AND, BIT_XOR, BIT_NOT, MODULO]
    )

    CONSTANTS = {
        "None": None,
        "True": True,
        "False": False,
    }
    def __init__(self, value, type=None, position=None):
        self.value = value
        self.type = type or value
        self.position = position

    def __str__(self):
        return f"{self.type} -> {self.value}"

class Lexer:
    def __init__(self, str):
        self.str = str
        self.pos = 0
        self.advance()

    def build_token(self, value, type=None):
        return Token(value, type, position=self.token_start_pos)

    @staticmethod
    def is_name_start(c):
        return c and (c.isalpha() or c == Token.UNDER or c == Token.AT)

    @staticmethod
    def is_name_part(c):
        return Lexer.is_name_start(c) or c.isdigit()

    def advance(self):
        if self.pos < len(self.str):
            self.current_char = self.str[self.pos]
            self.pos += 1
        else:
            self.current_char = ''

        return self.current_char

    def is_whitespace(self, c):
        if c:
            return c.isspace()

    def skip_whitespaces(self):
        while self.current_char.isspace():
            self.advance()

    def skip_comment(self):
        while self.current_char != Token.RCURCLY:
            self.advance()

        self.advance()

    def number_token(self):
        token = ''

        while self.current_char.isdigit():
            token += self.current_char
            self.advance()

        if self.current_char == Token.DOT:
            token += Token.DOT
            self.advance()

            while self.current_char.isdigit():
                token += self.current_char
                self.advance()

            return self.build_token(float(token), Token.FLOAT_CONST)
        else:
            return self.build_token(int(token), Token.INT_CONST)

    def name_token(self, prefix=None):
        token = (prefix or '') + self.current_char
        self.advance()

        while Lexer.is_name_part(self.current_char):
            token += self.current_char
            self.advance()

        if token in Token.KEYWORDS:
            return self.build_token(token)
        elif token in Token.CONSTANTS:
            return self.build_token(Token.CONSTANTS[token], Token.CONST)
        else:
            return self.build_token(token, Token.ID)

    def string_literal(self):
        token = ''
        self.advance()

        while self.current_char != Token.DQUOTE:
            token += self.current_char
            self.advance()

        self.advance()

        return self.build_token(token, Token.STR_CONST)

    def dollar_literal(self):
        self.advance()

        if self.current_char.isdigit():
            return self.number_token()
        else:
            return self.name_token(Token.DOLLAR)

    def colon_op(self):
        self.advance()

        if self.current_char == Token.EQ:
            self.advance()
            return self.build_token(Token.ASSIGN)

        return self.build_token(Token.COLON)

    def comparison_op(self):
        token = ''

        while self.current_char in Token.COMPARISON_SYMBOLS:
            token += self.current_char
            self.advance()

        # Temporary treat "="" as an "=="
        if token == Token.EQ:
            token = Token.EQ2

        return self.build_token(token)

    def div_op(self):
        self.advance()

        if self.current_char == Token.DIV:
            self.advance()
            return self.build_token(Token.INT_DIV)
        else:
            return self.build_token(Token.DIV)

    def mul_or_power_op(self):
        self.advance()

        if self.current_char == Token.MUL:
            self.advance()
            return self.build_token(Token.POWER)
        else:
            return self.build_token(Token.MUL)

    def symbol(self):
        sym = self.current_char
        self.advance()
        return self.build_token(sym)

    def next_token(self):
        self.skip_whitespaces()

        while self.current_char:
            if self.current_char == Token.LCURCLY:
                self.skip_comment()
                self.skip_whitespaces()
                continue

            self.token_start_pos = self.pos

            if self.current_char.isdigit():
                return self.number_token()

            if Lexer.is_name_start(self.current_char):
                return self.name_token()

            if self.current_char == Token.DQUOTE:
                return self.string_literal()

            if self.current_char == Token.DOLLAR:
                return self.dollar_literal()

            if self.current_char == Token.COLON:
                return self.colon_op()

            if self.current_char in Token.COMPARISON_SYMBOLS:
                return self.comparison_op()

            if self.current_char == Token.DIV:
                return self.div_op()

            if self.current_char == Token.MUL:
                return self.mul_or_power_op()

            if self.current_char in Token.SYMBOLS:
                return self.symbol()

            raise LexerError(f"unexpected char: '{self.current_char}' code={ord(self.current_char)} pos=#{self.pos}")

        return self.build_token(Token.EOF)

    def all_tokens(self):
        res = []
        token = self.next_token()

        while token.type != Token.EOF:
            res.append(token)
            token = self.next_token()

        return res



from datetime import datetime

from .cloud.experiment import AugerExperimentApi
from .exceptions import AugerException
from .cloud.experiment_session import AugerExperimentSessionApi


class Experiment(AugerExperimentApi):
    """Auger Cloud Experiments(s) management"""

    def __init__(self, ctx, dataset, experiment_name=None):
        super(Experiment, self).__init__(
            ctx, dataset.project, experiment_name)
        self.dataset = dataset

    def list(self):
        # data_set_id = self.dataset.oid
        # filter_by_dataset = \
        #     lambda exp: exp.get('project_file_id') == data_set_id
        # return (e for e in super().list() if filter_by_dataset(e))
        return super().list()

    def start(self):
        if self.dataset is None:
            raise AugerException(
                'DataSet is required to start Experiment...')

        if not self.dataset.is_exists:
            raise AugerException('Can\'t find DataSet on Auger Cloud...')

        if self.object_name and self.is_exists:
            data_set_id = self.dataset.oid
            experiment_data_set = self.properties().get('project_file_id')
            if data_set_id != experiment_data_set:
                self.object_id = None
                self.object_name = None
                self.ctx.log('Current experiment setup with different DataSet. '
                    'Will create new Experiment. data_set_id: %s, experiment_data_set: %s'%(data_set_id,experiment_data_set))
                
        self.dataset.project.start()

        if (self.object_name is None) or (not self.is_exists):
            self.create(self.dataset.name)
            self.ctx.log('Created Experiment %s ' % self.name)

        experiment_session_id = self.run()
        self.ctx.log('Started Experiment %s search...' % self.name)

        return self.name, experiment_session_id

    def stop(self, run_id=None):
        if run_id is None:
            run_id = self._get_latest_run()
            
        session_api = AugerExperimentSessionApi(
            self.ctx, None, None, run_id)
        return session_api.interrupt()

    def leaderboard(self, run_id=None):
        if run_id is None:
            run_id = self._get_latest_run()

        if run_id is None:
            return None, None, None
        else:
            session_api = AugerExperimentSessionApi(
                self.ctx, None, None, run_id)
            session_props = session_api.properties()
            status = session_props.get('status')
            errors = None
            if session_props.get('providers_data'):
                if session_props['providers_data'].get(self.ctx.get_name()):
                    errors = session_props['providers_data'][self.ctx.get_name()].get('errors')
            if not errors:
                errors = session_props.get('model_settings', {}).get('errors')

            return session_api.get_leaderboard(), status, run_id, \
                session_props.get('model_settings', {}).get('completed_evaluations', 0), errors
                
    def history(self):
        return AugerExperimentSessionApi(self.ctx, self).list()

    def _get_latest_run(self):
        latest = [None, None]
        for run in iter(self.history()):
            start_time = run.get('model_settings').get('start_time')
            if start_time:
                start_time = datetime.strptime(
                    start_time, '%Y-%m-%d %H:%M:%S.%f')
                if (latest[0] is None) or (latest[1] < start_time):
                    latest = [run.get('id'), start_time]
        return latest[0]

    def wait(self, run_id=None):
        if run_id is None:
            run_id = self._get_latest_run()

        if run_id is None:
            return None, None
        else:
            AugerExperimentSessionApi(
                self.ctx, self, None, run_id).wait_for_status([
                    'waiting', 'preprocess', 'started'])



import numpy as np
import pandas as pd
import os

from sklearn.mixture import GaussianMixture

from a2ml.api.model_review.model_helper import ModelHelper
from a2ml.api.model_review.model_review import ModelReview
from a2ml.api.utils import fsclient
from a2ml.api.utils.dataframe import DataFrame


class FeatureDivergence:
    NUMERIC_TYPES = {
        'integer',
        'double',
    }

    CATEGORICAL_TYPES = {
        'categorical',
        'hashing',
        'boolean',
    }

    IGNORED_TYPES = {
        'datetime',
        'timeseries',
    }

    MODEL_NAME = 'density_model.pkl'

    class Histogram:
        def fit(self, X, y=None):
            X = X.replace([np.inf, -np.inf], np.nan).dropna()
            X = np.copy(X)
            self.hist_ = []

            for i in range(X.shape[1]):
                values, counts = np.unique(X[:, i], return_counts=True)
                self.hist_.append(
                    {v: p for v, p in zip(values, counts / X.shape[0])})

            return self

        def score_samples(self, X, y=None):
            X = X.replace([np.inf, -np.inf], np.nan).dropna()
            X = np.copy(X)
            res = np.zeros_like(X, dtype=float)

            for i in range(X.shape[1]):
                for v, p in self.hist_[i].items():
                    res[X[:, i] == v, i] = p

            return np.sum(np.log(res), axis=1)

    class DensityEstimator:
        def __init__(self, num_cols: list = None, cat_cols: list = None):
            self.num_cols = num_cols or []
            self.cat_cols = cat_cols or []

        def fit(self, X, y=None):
            X = X.replace([np.inf, -np.inf], np.nan).dropna()

            self.cont_ = None
            self.cat_ = None

            if self.num_cols:
                self.cont_ = GaussianMixture(
                    n_components=3, covariance_type="diag").fit(X[self.num_cols])

            if self.cat_cols:
                self.cat_ = FeatureDivergence.Histogram().fit(X[self.cat_cols])

            return self

        def score(self, X, y=None):
            X = X.replace([np.inf, -np.inf], np.nan).dropna()
            num_score = self.cont_.score_samples(
                X[self.num_cols]) if self.cont_ is not None else np.zeros(X.shape[0])
            cat_score = self.cat_.score_samples(
                X[self.cat_cols]) if self.cat_ is not None else np.zeros(X.shape[0])

            return np.mean(num_score + cat_score)

    class DensityEstimatorPerFeature:
        def __init__(self, features, numerical_features, categorical_features):
            self.features = features
            self.numerical_features = set(numerical_features)
            self.categorical_features = set(categorical_features)
            self.models = {}
            self.base_values = {}

        def fit(self, df):
            for feature in self.features:
                model = None

                if feature in self.numerical_features:
                    model = FeatureDivergence.DensityEstimator([feature], [])

                if feature in self.categorical_features:
                    model = FeatureDivergence.DensityEstimator([], [feature])

                if model:
                    model.fit(df)
                    self.models[feature] = model
                    self.base_values[feature] = model.score(df)

        def score(self, feature, df):
            if feature in self.models:
                return self.models[feature].score(df) / self.base_values[feature]
            else:
                raise KeyError("Feature: '" + feature +
                               "'' is not in the model")

    class ModelIsNotReadyError(Exception):
        pass

    def __init__(self, params):
        self.params = params
        self.experiment_session = params['hub_info']['experiment_session']
        self.stat_data = self.experiment_session.get(
            'dataset_statistics', {}).get('stat_data', [])

    def build_and_save_model(self, divergence_model_name=None):
        df = DataFrame.create_dataframe(
            data_path=self.params['data_path'], features=self._get_density_features())

        model = self.DensityEstimatorPerFeature(
            self._get_density_features(),
            self._get_numerical_features(),
            self._get_categorical_features()
        )
        model.fit(df.df)

        path = self._get_divergence_model_path(divergence_model_name)
        fsclient.save_object_to_file(model, path)
        return path

    def score_divergence_daily(self, date_from=None, date_to=None, divergence_model_name=None, top_n=None):
        model_path = ModelHelper.get_model_path(params=self.params)
        div_model_path = self._get_divergence_model_path(divergence_model_name)

        if not fsclient.is_file_exists(div_model_path):
            raise self.ModelIsNotReadyError(div_model_path)

        divergence_model = fsclient.load_object_from_file(div_model_path)
        features = self._get_density_features()

        features_source = features
        feature_importances = self._get_most_important_features(
            features, top_n)

        if top_n:
            features_source = feature_importances.keys()

        res = {}

        for (curr_date, files) in ModelReview._prediction_files_by_day(
            model_path,
            date_from,
            date_to,
            "_*_actuals.feather.zstd"
        ):
            daily_df = None
            for (file, df) in DataFrame.load_from_files(files, features):
                if daily_df != None:
                    daily_df.df = pd.concat(
                        [daily_df.df, df.df], ignore_index=True)
                else:
                    daily_df = df

            if daily_df != None:
                sub_res = {}

                for feature in features_source:
                    sub_res[feature] = divergence_model.score(
                        feature, daily_df.df)

                res[str(curr_date)] = sub_res

        return {
            'divergence': res,
            'importance': feature_importances,
        }

    def _get_most_important_features(self, features, top_n):
        features_set = set(features)
        res = ModelReview(self.params).get_feature_importances()
        # Get feature importance for our features
        res = {key: res.get(key) for key in features_set}
        res = [[res[k], k] for k in res if res.get(k) != None]
        # Sort by importance desc
        res.sort(reverse=True)
        # Return top features by importance dict(feature => importance)
        return dict(map(lambda x: [x[1], x[0]], res[0:top_n]))

    def _get_divergence_model_path(self, divergence_model_name=None):
        experiment_session_path = ModelHelper.get_experiment_session_path(
            self.params)
        return os.path.join(experiment_session_path, divergence_model_name or self.MODEL_NAME)

    def _get_density_features(self):
        def predicate(col): return self._is_column_used(
            col) and not self._is_column_ignored(col)
        return [col['column_name'] for col in self.stat_data if predicate(col)]

    def _get_numerical_features(self):
        def predicate(col): return self._is_column_used(
            col) and self._is_column_numerical(col)
        return [col['column_name'] for col in self.stat_data if predicate(col)]

    def _get_categorical_features(self):
        def predicate(col): return self._is_column_used(
            col) and self._is_column_categorical(col)
        return [col['column_name'] for col in self.stat_data if predicate(col)]

    def _is_column_used(self, column):
        return column['isTarget'] or column['use']

    def _is_column_numerical(self, column):
        return column['datatype'] in self.NUMERIC_TYPES

    def _is_column_categorical(self, column):
        return column['datatype'] in self.CATEGORICAL_TYPES

    def _is_column_ignored(self, column):
        return column['datatype'] in self.IGNORED_TYPES
